import pytest

import numpy as np
from numpy.testing import assert_allclose

from quimb import (
    ham_heis,
    expec,
    plus,
    is_eigenvector,
    eigh,
    heisenberg_energy,
)

from quimb.tensor import (
    MPS_rand_state,
    MPS_product_state,
    MPS_computational_state,
    MPO_ham_ising,
    MPO_ham_XY,
    MPO_ham_heis,
    MPO_ham_mbl,
    MovingEnvironment,
    DMRG1,
    DMRG2,
    DMRGX,
    SpinHam1D,
)


np.random.seed(42)


class TestMovingEnvironment:
    def test_bsz1_start_left(self):
        tn = MPS_rand_state(6, bond_dim=7)
        env = MovingEnvironment(tn, begin='left', bsz=1)
        assert env.pos == 0
        assert len(env().tensors) == 3
        env.move_right()
        assert env.pos == 1
        assert len(env().tensors) == 3
        env.move_right()
        assert env.pos == 2
        assert len(env().tensors) == 3
        env.move_to(5)
        assert env.pos == 5
        assert len(env().tensors) == 3

    def test_bsz1_start_right(self):
        tn = MPS_rand_state(6, bond_dim=7)
        env = MovingEnvironment(tn, begin='right', bsz=1)
        assert env.pos == 5
        assert len(env().tensors) == 3
        env.move_left()
        assert env.pos == 4
        assert len(env().tensors) == 3
        env.move_left()
        assert env.pos == 3
        assert len(env().tensors) == 3
        env.move_to(0)
        assert env.pos == 0
        assert len(env().tensors) == 3

    def test_bsz2_start_left(self):
        tn = MPS_rand_state(6, bond_dim=7)
        env = MovingEnvironment(tn, begin='left', bsz=2)
        assert len(env().tensors) == 4
        env.move_right()
        assert len(env().tensors) == 4
        env.move_right()
        assert len(env().tensors) == 4
        with pytest.raises(ValueError):
            env.move_to(5)
        env.move_to(4)
        assert env.pos == 4
        assert len(env().tensors) == 4

    def test_bsz2_start_right(self):
        tn = MPS_rand_state(6, bond_dim=7)
        env = MovingEnvironment(tn, begin='right', bsz=2)
        assert env.pos == 4
        assert len(env().tensors) == 4
        env.move_left()
        assert env.pos == 3
        assert len(env().tensors) == 4
        env.move_left()
        assert env.pos == 2
        assert len(env().tensors) == 4
        with pytest.raises(ValueError):
            env.move_to(-1)
        env.move_to(0)
        assert env.pos == 0
        assert len(env().tensors) == 4

    @pytest.mark.parametrize("n", [20, 19])
    @pytest.mark.parametrize("bsz", [1, 2])
    @pytest.mark.parametrize("ssz", [1 / 2, 1.0])
    def test_cyclic_moving_env_init_left(self, n, bsz, ssz):
        nenv = 2
        p = MPS_rand_state(n, 4, cyclic=True)
        norm = p.H & p
        mes = MovingEnvironment(norm, begin='left', bsz=bsz,
                                cyclic=True, ssz=ssz)
        assert len(mes.envs) == n // 2 + n % 2
        assert mes.pos == 0
        assert len(mes.envs[0].tensors) == 2 * bsz + nenv
        assert len(mes.envs[n // 2 - 1].tensors) == 2 * bsz + 1
        assert n // 2 + n % 2 not in mes.envs
        assert n - 1 not in mes.envs

        for i in range(1, 2 * n):
            mes.move_right()
            assert mes.pos == i % n
            cur_env = mes()
            assert len(cur_env.tensors) == 2 * bsz + nenv
            assert (cur_env ^ all) == pytest.approx(1.0)

    @pytest.mark.parametrize("n", [20, 19])
    @pytest.mark.parametrize("bsz", [1, 2])
    @pytest.mark.parametrize("ssz", [1 / 2, 1.0])
    def test_cyclic_moving_env_init_right(self, n, bsz, ssz):
        p = MPS_rand_state(n, 4, cyclic=True)
        norm = p.H | p
        mes = MovingEnvironment(norm, begin='right', bsz=bsz,
                                cyclic=True, ssz=ssz)
        assert len(mes.envs) == n // 2 + n % 2
        assert mes.pos == n - 1
        assert len(mes.envs[n - 1].tensors) == 2 * bsz + 2
        assert len(mes.envs[n - n // 2].tensors) == 2 * bsz + 1
        assert 0 not in mes.envs
        assert n // 2 - 1 not in mes.envs

        for i in reversed(range(-n, n - 1)):
            mes.move_left()
            assert mes.pos == i % n
            cur_env = mes()
            assert len(cur_env.tensors) == 2 * bsz + 2
            assert (cur_env ^ all) == pytest.approx(1.0)


class TestDMRG1:

    def test_single_explicit_sweep(self):
        h = MPO_ham_heis(5)
        dmrg = DMRG1(h, bond_dims=3)
        assert dmrg._k[0].dtype == float

        energy_tn = (dmrg._b | dmrg.ham | dmrg._k)

        e0 = energy_tn ^ ...
        assert abs(e0.imag) < 1e-13

        de1 = dmrg.sweep_right()
        e1 = energy_tn ^ ...
        assert_allclose(de1, e1)
        assert abs(e1.imag) < 1e-13

        de2 = dmrg.sweep_right()
        e2 = energy_tn ^ ...
        assert_allclose(de2, e2)
        assert abs(e2.imag) < 1e-13

        # state is already left canonized after right sweep
        de3 = dmrg.sweep_left(canonize=False)
        e3 = energy_tn ^ ...
        assert_allclose(de3, e3)
        assert abs(e2.imag) < 1e-13

        de4 = dmrg.sweep_left()
        e4 = energy_tn ^ ...
        assert_allclose(de4, e4)
        assert abs(e2.imag) < 1e-13

        # test still normalized
        assert dmrg._k[0].dtype == float
        dmrg._k.align_(dmrg._b)
        assert_allclose(abs(dmrg._b @ dmrg._k), 1)

        assert e1.real < e0.real
        assert e2.real < e1.real
        assert e3.real < e2.real
        assert e4.real < e3.real

    @pytest.mark.parametrize("dense", [False, True])
    @pytest.mark.parametrize("MPO_ham", [MPO_ham_XY, MPO_ham_heis])
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_ground_state_matches(self, dense, MPO_ham, cyclic):
        n = 10

        tol = 3e-2 if cyclic else 1e-4

        h = MPO_ham(n, cyclic=cyclic)
        dmrg = DMRG1(h, bond_dims=[4, 8, 12])
        dmrg.opts['local_eig_ham_dense'] = dense
        dmrg.opts['periodic_segment_size'] = 1.0
        dmrg.opts['periodic_nullspace_fudge_factor'] = 1e-6
        assert dmrg.solve(tol=tol / 10, verbosity=1)
        assert dmrg.state.cyclic == cyclic
        eff_e, mps_gs = dmrg.energy, dmrg.state
        mps_gs_dense = mps_gs.to_dense()

        assert_allclose(mps_gs_dense.H @ mps_gs_dense, 1.0, rtol=tol)

        h_dense = h.to_dense()

        # check against dense form
        actual_e, gs = eigh(h_dense, k=1)
        assert_allclose(actual_e, eff_e, rtol=tol)
        assert_allclose(abs(expec(mps_gs_dense, gs)), 1.0, rtol=tol)

        # check against actual MPO_ham
        if MPO_ham is MPO_ham_XY:
            ham_dense = ham_heis(n, cyclic=cyclic,
                                 j=(1.0, 1.0, 0.0), sparse=True)
        elif MPO_ham is MPO_ham_heis:
            ham_dense = ham_heis(n, cyclic=cyclic, sparse=True)

        actual_e, gs = eigh(ham_dense, k=1)
        assert_allclose(actual_e, eff_e, rtol=tol)
        assert_allclose(abs(expec(mps_gs_dense, gs)), 1.0, rtol=tol)

    def test_ising_and_MPS_product_state(self):
        h = MPO_ham_ising(6, bx=2.0, j=0.1)
        dmrg = DMRG1(h, bond_dims=8)
        assert dmrg.solve(verbosity=1)
        eff_e, mps_gs = dmrg.energy, dmrg.state
        mps_gs_dense = mps_gs.to_dense()
        assert_allclose(mps_gs_dense.H @ mps_gs_dense, 1.0)

        # check against dense
        h_dense = h.to_dense()
        actual_e, gs = eigh(h_dense, k=1)
        assert_allclose(actual_e, eff_e)
        assert_allclose(abs(expec(mps_gs_dense, gs)), 1.0)

        exp_gs = MPS_product_state([plus()] * 6)
        assert_allclose(abs(exp_gs.H @ mps_gs), 1.0, rtol=1e-3)


class TestDMRG2:
    @pytest.mark.parametrize("dense", [False, True])
    @pytest.mark.parametrize("MPO_ham", [MPO_ham_XY, MPO_ham_heis])
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_matches_exact(self, dense, MPO_ham, cyclic):
        n = 6
        h = MPO_ham(n, cyclic=cyclic)

        tol = 3e-2 if cyclic else 1e-4

        dmrg = DMRG2(h, bond_dims=[4, 8, 12])
        assert dmrg._k[0].dtype == float
        dmrg.opts['local_eig_ham_dense'] = dense
        dmrg.opts['periodic_segment_size'] = 1.0
        dmrg.opts['periodic_nullspace_fudge_factor'] = 1e-6

        assert dmrg.solve(tol=tol / 10, verbosity=1)

        # XXX: need to dispatch SLEPc eigh on real input
        # assert dmrg._k[0].dtype == float

        eff_e, mps_gs = dmrg.energy, dmrg.state
        mps_gs_dense = mps_gs.to_dense()

        assert_allclose(expec(mps_gs_dense, mps_gs_dense), 1.0, rtol=tol)

        h_dense = h.to_dense()

        # check against dense form
        actual_e, gs = eigh(h_dense, k=1)
        assert_allclose(actual_e, eff_e, rtol=tol)
        assert_allclose(abs(expec(mps_gs_dense, gs)), 1.0, rtol=tol)

        # check against actual MPO_ham
        if MPO_ham is MPO_ham_XY:
            ham_dense = ham_heis(n, cyclic=cyclic, j=(1.0, 1.0, 0.0))
        elif MPO_ham is MPO_ham_heis:
            ham_dense = ham_heis(n, cyclic=cyclic)

        actual_e, gs = eigh(ham_dense, k=1)
        assert_allclose(actual_e, eff_e, rtol=tol)
        assert_allclose(abs(expec(mps_gs_dense, gs)), 1.0, rtol=tol)

    def test_cyclic_solve_big_with_segmenting(self):
        n = 150
        ham = MPO_ham_heis(n, cyclic=True)
        dmrg = DMRG2(ham, bond_dims=range(10, 30, 2))
        dmrg.opts['periodic_segment_size'] = 1 / 3
        assert dmrg.solve(tol=1, verbosity=2)
        assert dmrg.energy == pytest.approx(heisenberg_energy(n), 1e-3)

    @pytest.mark.parametrize("dtype", [np.float32, np.float64,
                                       np.complex64, np.complex128])
    def test_dtypes(self, dtype):
        H = MPO_ham_heis(8).astype(dtype)
        dmrg = DMRG2(H)
        dmrg.opts['local_eig_backend'] = 'scipy'
        dmrg.solve(max_sweeps=3)
        res_dtype, = {t.dtype for t in dmrg.state}
        assert res_dtype == dtype

    def test_total_size_2(self):
        N = 2
        builder = SpinHam1D(1 / 2)
        for i in range(N - 1):
            builder[i, i + 1] += 1.0, 'Z', 'Z'

        H = builder.build_mpo(N)
        dmrg = DMRG2(H)
        dmrg.solve(verbosity=1)
        assert dmrg.energy == pytest.approx(-1 / 4)

    def test_variable_bond_ham(self):
        import quimb as qu

        HB = SpinHam1D(1 / 2)
        HB[0, 1] += 0.6, 'Z', 'Z'
        HB[1, 2] += 0.7, 'Z', 'Z'
        HB[1, 2] += 0.8, 'X', 'X'
        HB[2, 3] += 0.9, 'Y', 'Y'

        H_mpo = HB.build_mpo(4)
        H_sps = HB.build_sparse(4)

        assert H_mpo.bond_sizes() == [3, 4, 3]

        Sx, Sy, Sz = map(qu.spin_operator, 'xyz')
        H_explicit = (
            qu.ikron(0.6 * Sz & Sz, [2, 2, 2, 2], [0, 1]) +
            qu.ikron(0.7 * Sz & Sz, [2, 2, 2, 2], [1, 2]) +
            qu.ikron(0.8 * Sx & Sx, [2, 2, 2, 2], [1, 2]) +
            qu.ikron(0.9 * Sy & Sy, [2, 2, 2, 2], [2, 3])
        )

        assert_allclose(H_explicit, H_mpo.to_dense())
        assert_allclose(H_explicit, H_sps.A)


class TestDMRGX:

    def test_explicit_sweeps(self):
        n = 8
        chi = 16
        ham = MPO_ham_mbl(n, dh=4, seed=42)
        p0 = MPS_rand_state(n, 2).expand_bond_dimension(chi)

        b0 = p0.H
        p0.align_(ham, b0)
        en0 = (p0 & ham & b0) ^ ...
        dmrgx = DMRGX(ham, p0, chi)
        dmrgx.sweep_right()
        en1 = dmrgx.sweep_left(canonize=False)
        assert en0 != en1

        dmrgx.sweep_right(canonize=False)
        en = dmrgx.sweep_right(canonize=True)

        # check normalized
        assert_allclose(dmrgx._k.H @ dmrgx._k, 1.0)

        k = dmrgx._k.to_dense()
        h = ham.to_dense()
        el, ev = eigh(h)

        # check variance very low
        assert np.abs((k.H @ h @ h @ k) - (k.H @ h @ k)**2) < 1e-12

        # check exactly one eigenvalue matched well
        assert np.sum(np.abs(el - en) < 1e-12) == 1

        # check exactly one eigenvector is matched with high fidelity
        ovlps = (ev.H @ k).A**2
        big_ovlps = ovlps[ovlps > 1e-12]
        assert_allclose(big_ovlps, [1])

        # check fully
        assert is_eigenvector(k, h, tol=1e-10)

    def test_solve_bigger(self):
        n = 14
        chi = 16
        ham = MPO_ham_mbl(n, dh=8, seed=42)
        p0 = MPS_computational_state('00110111000101')
        dmrgx = DMRGX(ham, p0, chi)
        assert dmrgx.solve(tol=1e-5, sweep_sequence='R')
        assert dmrgx.state[0].dtype == float



from itertools import starmap

import numpy as np
import scipy.sparse.linalg as spla
from opt_einsum import shared_intermediates
from autoray import do, dag, conj, reshape

from ..utils import pairwise
from .drawing import get_colors
from .tensor_core import Tensor, contract_strategy
from .optimize import TNOptimizer
from .tensor_2d import (
    gen_2d_bonds,
    calc_plaquette_sizes,
    calc_plaquette_map,
    plaquette_to_sites,
    gen_long_range_path,
    gen_long_range_swap_path,
    swap_path_to_long_range_path,
    nearest_neighbors,
)
from .tensor_arbgeom_tebd import LocalHamGen, TEBDGen


class LocalHam2D(LocalHamGen):
    """A 2D Hamiltonian represented as local terms. This combines all two site
    and one site terms into a single interaction per lattice pair, and caches
    operations on the terms such as getting their exponential.

    Parameters
    ----------
    Lx : int
        The number of rows.
    Ly : int
        The number of columns.
    H2 : array_like or dict[tuple[tuple[int]], array_like]
        The two site term(s). If a single array is given, assume to be the
        default interaction for all nearest neighbours. If a dict is supplied,
        the keys should represent specific pairs of coordinates like
        ``((ia, ja), (ib, jb))`` with the values the array representing the
        interaction for that pair. A default term for all remaining nearest
        neighbours interactions can still be supplied with the key ``None``.
    H1 : array_like or dict[tuple[int], array_like], optional
        The one site term(s). If a single array is given, assume to be the
        default onsite term for all terms. If a dict is supplied,
        the keys should represent specific coordinates like
        ``(i, j)`` with the values the array representing the local term for
        that site. A default term for all remaining sites can still be supplied
        with the key ``None``.

    Attributes
    ----------
    terms : dict[tuple[tuple[int]], array_like]
        The total effective local term for each interaction (with single site
        terms appropriately absorbed). Each key is a pair of coordinates
        ``ija, ijb`` with ``ija < ijb``.

    """

    def __init__(self, Lx, Ly, H2, H1=None):
        self.Lx = int(Lx)
        self.Ly = int(Ly)

        # parse two site terms
        if hasattr(H2, 'shape'):
            # use as default nearest neighbour term
            H2 = {None: H2}
        else:
            H2 = dict(H2)

        # possibly fill in default gates
        default_H2 = H2.pop(None, None)
        if default_H2 is not None:
            for coo_a, coo_b in gen_2d_bonds(Lx, Ly, steppers=[
                lambda i, j: (i, j + 1),
                lambda i, j: (i + 1, j),
            ]):
                if (coo_a, coo_b) not in H2 and (coo_b, coo_a) not in H2:
                    H2[coo_a, coo_b] = default_H2

        super().__init__(H2=H2, H1=H1)

    @property
    def nsites(self):
        """The number of sites in the system.
        """
        return self.Lx * self.Ly

    def __repr__(self):
        s = "<LocalHam2D(Lx={}, Ly={}, num_terms={})>"
        return s.format(self.Lx, self.Ly, len(self.terms))

    def draw(
        self,
        ordering='sort',
        show_norm=True,
        figsize=None,
        fontsize=8,
        legend=True,
        ax=None,
        return_fig=False,
        **kwargs,
    ):
        """Plot this Hamiltonian as a network.

        Parameters
        ----------
        ordering : {'sort', None, 'random'}, optional
            An ordering of the termns, or an argument to be supplied to
            :meth:`quimb.tensor.tensor_2d_tebd.LocalHam2D.get_auto_ordering`
            to generate this automatically.
        show_norm : bool, optional
            Show the norm of each term as edge labels.
        figsize : None or tuple[int], optional
            Size of the figure, defaults to size of Hamiltonian.
        fontsize : int, optional
            Font size for norm labels.
        legend : bool, optional
            Whether to show the legend of which terms are in which group.
        ax : None or matplotlib.Axes, optional
            Add to a existing set of axes.
        return_fig : bool, optional
            Whether to return any newly created figure.
        """
        import matplotlib.pyplot as plt

        if figsize is None:
            figsize = (self.Ly, self.Lx)

        ax_supplied = (ax is not None)
        if not ax_supplied:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            ax.axis('off')
            ax.set_aspect('equal')

        if ordering is None or isinstance(ordering, str):
            ordering = self.get_auto_ordering(ordering, **kwargs)

        data = []
        seen = set()
        n = 0
        for ij1, ij2 in ordering:
            if (ij1 in seen) or (ij2 in seen):
                # start a new group
                seen = {ij1, ij2}
                n += 1
            else:
                seen.add(ij1)
                seen.add(ij2)

            ys, xs = zip(ij1, ij2)

            d = ((xs[1] - xs[0])**2 + (ys[1] - ys[0])**2)**0.5
            # offset by the length of bond to distinguish NNN etc.
            #     choose offset direction by parity of first site

            if d > 2**0.5:
                xs = [xi + (-1)**int(ys[0]) * 0.02 * d for xi in xs]
                ys = [yi + (-1)**int(xs[0]) * 0.02 * d for yi in ys]

            # set coordinates for label with some offset towards left
            if ij1[1] < ij2[1]:
                lbl_x0 = (3 * xs[0] + 2 * xs[1]) / 5
                lbl_y0 = (3 * ys[0] + 2 * ys[1]) / 5
            else:
                lbl_x0 = (2 * xs[0] + 3 * xs[1]) / 5
                lbl_y0 = (2 * ys[0] + 3 * ys[1]) / 5

            nrm = do('linalg.norm', self.terms[ij1, ij2])

            data.append((xs, ys, n, lbl_x0, lbl_y0, nrm))

        num_groups = n + 1
        colors = get_colors(range(num_groups))

        # do the plotting
        for xs, ys, n, lbl_x0, lbl_y0, nrm in data:
            ax.plot(xs, ys, c=colors[n], linewidth=2 * nrm**0.5)
            if show_norm:
                label = "{:.3f}".format(nrm)
                ax.text(lbl_x0, lbl_y0, label, c=colors[n], fontsize=fontsize)

        # create legend
        if legend:
            handles = []
            for color in colors.values():
                handles += [plt.Line2D([0], [0], marker='o', color=color,
                                       linestyle='', markersize=10)]

            lbls = [f"Group {i + 1}" for i in range(num_groups)]

            ax.legend(handles, lbls, ncol=max(round(len(handles) / 20), 1),
                      loc='center left', bbox_to_anchor=(1, 0.5))

        if ax_supplied:
            return

        if return_fig:
            return fig

        plt.show()

    graph = draw


class TEBD2D(TEBDGen):
    """Generic class for performing two dimensional time evolving block
    decimation, i.e. applying the exponential of a Hamiltonian using
    a product formula that involves applying local exponentiated gates only.

    Parameters
    ----------
    psi0 : TensorNetwork2DVector
        The initial state.
    ham : LocalHam2D
        The Hamtiltonian consisting of local terms.
    tau : float, optional
        The default local exponent, if considered as time real values here
        imply imaginary time.
    max_bond : {'psi0', int, None}, optional
        The maximum bond dimension to keep when applying each gate.
    gate_opts : dict, optional
        Supplied to :meth:`quimb.tensor.tensor_2d.TensorNetwork2DVector.gate`,
        in addition to ``max_bond``. By default ``contract`` is set to
        'reduce-split' and ``cutoff`` is set to ``0.0``.
    ordering : str, tuple[tuple[int]], callable, optional
        How to order the terms, if a string is given then use this as the
        strategy given to
        :meth:`~quimb.tensor.tensor_2d_tebd.LocalHam2D.get_auto_ordering`. An
        explicit list of coordinate pairs can also be given. The default is to
        greedily form an 'edge coloring' based on the sorted list of
        Hamiltonian pair coordinates. If a callable is supplied it will be used
        to generate the ordering before each sweep.
    second_order_reflect : bool, optional
        If ``True``, then apply each layer of gates in ``ordering`` forward
        with half the time step, then the same with reverse order.
    compute_energy_every : None or int, optional
        How often to compute and record the energy. If a positive integer 'n',
        the energy is computed *before* every nth sweep (i.e. including before
        the zeroth).
    compute_energy_final : bool, optional
        Whether to compute and record the energy at the end of the sweeps
        regardless of the value of ``compute_energy_every``. If you start
        sweeping again then this final energy is the same as the zeroth of the
        next set of sweeps and won't be recomputed.
    compute_energy_opts : dict, optional
        Supplied to
        :meth:`~quimb.tensor.tensor_2d.PEPS.compute_local_expectation`. By
        default ``max_bond`` is set to ``max(8, D**2)`` where ``D`` is the
        maximum bond to use for applying the gate, ``cutoff`` is set to ``0.0``
        and ``normalized`` is set to ``True``.
    compute_energy_fn : callable, optional
        Supply your own function to compute the energy, it should take the
        ``TEBD2D`` object as its only argument.
    callback : callable, optional
        A custom callback to run after every sweep, it should take the
        ``TEBD2D`` object as its only argument. If it returns any value
        that boolean evaluates to ``True`` then terminal the evolution.
    progbar : boolean, optional
        Whether to show a live progress bar during the evolution.
    kwargs
        Extra options for the specific ``TEBD2D`` subclass.

    Attributes
    ----------
    state : TensorNetwork2DVector
        The current state.
    ham : LocalHam2D
        The Hamiltonian being used to evolve.
    energy : float
        The current of the current state, this will trigger a computation if
        the energy at this iteration hasn't been computed yet.
    energies : list[float]
        The energies that have been computed, if any.
    its : list[int]
        The corresponding sequence of iteration numbers that energies have been
        computed at.
    taus : list[float]
        The corresponding sequence of time steps that energies have been
        computed at.
    best : dict
        If ``keep_best`` was set then the best recorded energy and the
        corresponding state that was computed - keys ``'energy'`` and
        ``'state'`` respectively.
    """

    def __init__(
        self,
        psi0,
        ham,
        tau=0.01,
        D=None,
        chi=None,
        imag=True,
        gate_opts=None,
        ordering=None,
        second_order_reflect=False,
        compute_energy_every=None,
        compute_energy_final=True,
        compute_energy_opts=None,
        compute_energy_fn=None,
        compute_energy_per_site=False,
        callback=None,
        keep_best=False,
        progbar=True,
    ):
        super().__init__(
            psi0=psi0,
            ham=ham,
            tau=tau,
            D=D,
            imag=imag,
            gate_opts=gate_opts,
            ordering=ordering,
            second_order_reflect=second_order_reflect,
            compute_energy_every=compute_energy_every,
            compute_energy_final=compute_energy_final,
            compute_energy_opts=compute_energy_opts,
            compute_energy_fn=compute_energy_fn,
            compute_energy_per_site=compute_energy_per_site,
            callback=callback,
            keep_best=keep_best,
            progbar=progbar,
        )

        # parse energy computation options
        if chi is None:
            chi = max(8, self.D**2)
        self.compute_energy_opts['max_bond'] = chi
        self.compute_energy_opts.setdefault('cutoff', 0.0)
        self.compute_energy_opts.setdefault('normalized', True)

    def compute_energy(self):
        """Compute and return the energy of the current state.
        """
        return self.state.compute_local_expectation(
            self.ham.terms,
            **self.compute_energy_opts
        )

    @property
    def chi(self):
        return self.compute_energy_opts['max_bond']

    @chi.setter
    def chi(self, value):
        self.compute_energy_opts['max_bond'] = round(value)

    def __repr__(self):
        s = "<{}(n={}, tau={}, D={}, chi={})>"
        return s.format(
            self.__class__.__name__, self.n, self.tau, self.D, self.chi)


def conditioner(tn, value=None, sweeps=2, balance_bonds=True):
    """
    """
    if balance_bonds:
        for _ in range(sweeps - 1):
            tn.balance_bonds_()
            tn.equalize_norms_()
        tn.balance_bonds_()
    tn.equalize_norms_(value=value)


class SimpleUpdate(TEBD2D):
    """A simple subclass of ``TEBD2D`` that overrides two key methods in
    order to keep 'diagonal gauges' living on the bonds of a PEPS. The gauges
    are stored separately from the main PEPS in the ``gauges`` attribute.
    Before and after a gate is applied they are absorbed and then extracted.
    When accessing the ``state`` attribute they are automatically inserted or
    you can call ``get_state(absorb_gauges=False)`` to lazily add them as
    hyperedge weights only. Reference: https://arxiv.org/abs/0806.3719.

    Parameters
    ----------
    psi0 : TensorNetwork2DVector
        The initial state.
    ham : LocalHam2D
        The Hamtiltonian consisting of local terms.
    tau : float, optional
        The default local exponent, if considered as time real values here
        imply imaginary time.
    max_bond : {'psi0', int, None}, optional
        The maximum bond dimension to keep when applying each gate.
    gate_opts : dict, optional
        Supplied to :meth:`quimb.tensor.tensor_2d.TensorNetwork2DVector.gate`,
        in addition to ``max_bond``. By default ``contract`` is set to
        'reduce-split' and ``cutoff`` is set to ``0.0``.
    ordering : str, tuple[tuple[int]], callable, optional
        How to order the terms, if a string is given then use this as the
        strategy given to
        :meth:`~quimb.tensor.tensor_2d_tebd.LocalHam2D.get_auto_ordering`. An
        explicit list of coordinate pairs can also be given. The default is to
        greedily form an 'edge coloring' based on the sorted list of
        Hamiltonian pair coordinates. If a callable is supplied it will be used
        to generate the ordering before each sweep.
    second_order_reflect : bool, optional
        If ``True``, then apply each layer of gates in ``ordering`` forward
        with half the time step, then the same with reverse order.
    compute_energy_every : None or int, optional
        How often to compute and record the energy. If a positive integer 'n',
        the energy is computed *before* every nth sweep (i.e. including before
        the zeroth).
    compute_energy_final : bool, optional
        Whether to compute and record the energy at the end of the sweeps
        regardless of the value of ``compute_energy_every``. If you start
        sweeping again then this final energy is the same as the zeroth of the
        next set of sweeps and won't be recomputed.
    compute_energy_opts : dict, optional
        Supplied to
        :meth:`~quimb.tensor.tensor_2d.PEPS.compute_local_expectation`. By
        default ``max_bond`` is set to ``max(8, D**2)`` where ``D`` is the
        maximum bond to use for applying the gate, ``cutoff`` is set to ``0.0``
        and ``normalized`` is set to ``True``.
    compute_energy_fn : callable, optional
        Supply your own function to compute the energy, it should take the
        ``TEBD2D`` object as its only argument.
    callback : callable, optional
        A custom callback to run after every sweep, it should take the
        ``TEBD2D`` object as its only argument. If it returns any value
        that boolean evaluates to ``True`` then terminal the evolution.
    progbar : boolean, optional
        Whether to show a live progress bar during the evolution.
    gauge_renorm : bool, optional
        Whether to actively renormalize the singular value gauges.
    gauge_smudge : float, optional
        A small offset to use when applying the guage and its inverse to avoid
        numerical problems.
    condition_tensors : bool, optional
        Whether to actively equalize tensor norms for numerical stability.
    condition_balance_bonds : bool, optional
        If and when equalizing tensor norms, whether to also balance bonds as
        an additional conditioning.
    long_range_use_swaps : bool, optional
        If there are long range terms, whether to use swap gates to apply the
        terms. If ``False``, a long range blob tensor (which won't scale well
        for long distances) is formed instead.
    long_range_path_sequence : str or callable, optional
        If there are long range terms how to generate the path between the two
        coordinates. If callable, should take the two coordinates and return a
        sequence of  coordinates that links them, else passed to
        ``gen_long_range_swap_path``.

    Attributes
    ----------
    state : TensorNetwork2DVector
        The current state.
    ham : LocalHam2D
        The Hamiltonian being used to evolve.
    energy : float
        The current of the current state, this will trigger a computation if
        the energy at this iteration hasn't been computed yet.
    energies : list[float]
        The energies that have been computed, if any.
    its : list[int]
        The corresponding sequence of iteration numbers that energies have been
        computed at.
    taus : list[float]
        The corresponding sequence of time steps that energies have been
        computed at.
    best : dict
        If ``keep_best`` was set then the best recorded energy and the
        corresponding state that was computed - keys ``'energy'`` and
        ``'state'`` respectively.
    """

    def __init__(
        self,
        psi0,
        ham,
        tau=0.01,
        D=None,
        chi=None,
        gauge_renorm=True,
        gauge_smudge=1e-6,
        condition_tensors=True,
        condition_balance_bonds=True,
        long_range_use_swaps=False,
        long_range_path_sequence='random',
        imag=True,
        gate_opts=None,
        ordering=None,
        second_order_reflect=False,
        compute_energy_every=None,
        compute_energy_final=True,
        compute_energy_opts=None,
        compute_energy_fn=None,
        compute_energy_per_site=False,
        callback=None,
        keep_best=False,
        progbar=True,
    ):
        super().__init__(
            psi0=psi0,
            ham=ham,
            tau=tau,
            D=D,
            chi=chi,
            imag=imag,
            gate_opts=gate_opts,
            ordering=ordering,
            second_order_reflect=second_order_reflect,
            compute_energy_every=compute_energy_every,
            compute_energy_final=compute_energy_final,
            compute_energy_opts=compute_energy_opts,
            compute_energy_fn=compute_energy_fn,
            compute_energy_per_site=compute_energy_per_site,
            callback=callback,
            keep_best=keep_best,
            progbar=progbar,
        )
        self.gauge_renorm = gauge_renorm
        self.gauge_smudge = gauge_smudge
        self.condition_tensors = condition_tensors
        self.condition_balance_bonds = condition_balance_bonds
        self.gate_opts['long_range_use_swaps'] = long_range_use_swaps
        self.long_range_path_sequence = long_range_path_sequence

    def _initialize_gauges(self):
        """Create unit singular values, stored as tensors.
        """
        # create the gauges like whatever data array is in the first site.
        data00 = next(iter(self._psi.tensor_map.values())).data

        self._gauges = dict()
        for ija, ijb in self._psi.gen_bond_coos():
            bnd = self._psi.bond(ija, ijb)
            d = self._psi.ind_size(bnd)
            Tsval = Tensor(
                do('ones', (d,), dtype=data00.dtype, like=data00),
                inds=[bnd],
                tags=[
                    self._psi.site_tag(*ija),
                    self._psi.site_tag(*ijb),
                    'SU_gauge',
                ]
            )
            self._gauges[tuple(sorted((ija, ijb)))] = Tsval

    @property
    def gauges(self):
        """The dictionary of bond pair coordinates to Tensors describing the
        weights (``t = gauges[pair]; t.data``) and index
        (``t = gauges[pair]; t.inds[0]``) of all the gauges.
        """
        return self._gauges

    @property
    def long_range_use_swaps(self):
        return self.gate_opts['long_range_use_swaps']

    @long_range_use_swaps.setter
    def long_range_use_swaps(self, b):
        self.gate_opts['long_range_use_swaps'] = bool(b)

    def gate(self, U, where):
        """Like ``TEBD2D.gate`` but absorb and extract the relevant gauges
        before and after each gate application.
        """
        ija, ijb = where

        if callable(self.long_range_path_sequence):
            long_range_path_sequence = self.long_range_path_sequence(ija, ijb)
        else:
            long_range_path_sequence = self.long_range_path_sequence

        if self.long_range_use_swaps:
            path = tuple(gen_long_range_swap_path(
                ija, ijb, sequence=long_range_path_sequence))
            string = swap_path_to_long_range_path(path, ija)
        else:
            # get the string linking the two sites
            string = path = tuple(gen_long_range_path(
                ija, ijb, sequence=long_range_path_sequence))

        def env_neighbours(i, j):
            return tuple(filter(
                lambda coo: self._psi.valid_coo((coo)) and coo not in string,
                nearest_neighbors((i, j))
            ))

        # get the relevant neighbours for string of sites
        neighbours = {site: env_neighbours(*site) for site in string}

        # absorb the 'outer' gauges from these neighbours
        for site in string:
            Tij = self._psi[site]
            for neighbour in neighbours[site]:
                Tsval = self.gauges[tuple(sorted((site, neighbour)))]
                Tij.multiply_index_diagonal_(
                    ind=Tsval.inds[0], x=(Tsval.data + self.gauge_smudge))

        # absorb the inner bond gauges equally into both sites along string
        for site_a, site_b in pairwise(string):
            Ta, Tb = self._psi[site_a], self._psi[site_b]
            Tsval = self.gauges[tuple(sorted((site_a, site_b)))]
            bnd, = Tsval.inds
            Ta.multiply_index_diagonal_(ind=bnd, x=Tsval.data**0.5)
            Tb.multiply_index_diagonal_(ind=bnd, x=Tsval.data**0.5)

        # perform the gate, retrieving new bond singular values
        info = dict()
        self._psi.gate_(U, where, absorb=None, info=info,
                        long_range_path_sequence=path, **self.gate_opts)

        # set the new singualar values all along the chain
        for site_a, site_b in pairwise(string):
            bond_pair = tuple(sorted((site_a, site_b)))
            s = info['singular_values', bond_pair]
            if self.gauge_renorm:
                # keep the singular values from blowing up
                s = s / s[0]
            Tsval = self.gauges[bond_pair]
            Tsval.modify(data=s)

        # absorb the 'outer' gauges from these neighbours
        for site in string:
            Tij = self._psi[site]
            for neighbour in neighbours[site]:
                Tsval = self.gauges[tuple(sorted((site, neighbour)))]
                Tij.multiply_index_diagonal_(
                    ind=Tsval.inds[0], x=(Tsval.data + self.gauge_smudge)**-1)

    def get_state(self, absorb_gauges=True):
        """Return the state, with the diagonal bond gauges either absorbed
        equally into the tensors on either side of them
        (``absorb_gauges=True``, the default), or left lazily represented in
        the tensor network with hyperedges (``absorb_gauges=False``).
        """
        psi = self._psi.copy()

        if not absorb_gauges:
            for Tsval in self.gauges.values():
                psi &= Tsval
        else:
            for (ija, ijb), Tsval in self.gauges.items():
                bnd, = Tsval.inds
                Ta = psi[ija]
                Tb = psi[ijb]
                Ta.multiply_index_diagonal_(bnd, Tsval.data**0.5)
                Tb.multiply_index_diagonal_(bnd, Tsval.data**0.5)

        if self.condition_tensors:
            conditioner(psi, balance_bonds=self.condition_balance_bonds)

        return psi

    def set_state(self, psi):
        """Set the wavefunction state, this resets the environment gauges to
        unity.
        """
        self._psi = psi.copy()
        self._initialize_gauges()


def gate_full_update_als(
    ket,
    env,
    bra,
    G,
    where,
    tags_plq,
    steps,
    tol,
    max_bond,
    optimize='auto-hq',
    solver='solve',
    dense=True,
    enforce_pos=False,
    pos_smudge=1e-6,
    init_simple_guess=True,
    condition_tensors=True,
    condition_maintain_norms=True,
    condition_balance_bonds=True,
):
    ket_plq = ket.select_any(tags_plq)
    bra_plq = bra.select_any(tags_plq)

    # this is the full target (copy - not virtual)
    target = ket_plq.gate(G, where, contract=False) | env

    if init_simple_guess:
        ket_plq.gate_(G, where, contract='reduce-split', max_bond=max_bond)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))

    if condition_tensors:
        conditioner(ket_plq, balance_bonds=condition_balance_bonds)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))
        if condition_maintain_norms:
            pre_norm = ket_plq[site].norm()

    overlap = bra_plq | target
    norm_plq = bra_plq | env | ket_plq

    xs = dict()
    x_previous = dict()
    previous_cost = None

    with contract_strategy(optimize), shared_intermediates():
        for i in range(steps):

            for site in tags_plq:
                lix = norm_plq[site, 'BRA'].inds[:-1]
                rix = norm_plq[site, 'KET'].inds[:-1]
                # remove site tensors and group their indices
                if dense:
                    N = (norm_plq.select(site, which='!any')
                         .to_dense(lix, rix))

                    if enforce_pos:
                        el, ev = do('linalg.eigh', (N + dag(N)) / 2)
                        el = do('clip', el, pos_smudge, None)
                        N = ev @ do('diag', el) @ dag(ev)

                else:
                    N = (norm_plq.select(site, which='!any')
                         .aslinearoperator(lix, rix))

                # target vector (remove lower site tensor and contract to vec)
                b = (overlap
                     .select((site, 'BRA'), which='!all')
                     .to_dense(overlap[site, 'BRA'].inds[:-1],
                               overlap[site, 'BRA'].inds[-1:]))

                if solver == 'solve':
                    x = do('linalg.solve', N, b)
                elif solver == 'lstsq':
                    x = do('linalg.lstsq', N, b, rcond=tol * 1e-3)[0]
                else:
                    # use scipy sparse linalg solvers
                    if solver in ('lsqr', 'lsmr'):
                        solver_opts = dict(atol=tol, btol=tol)
                    else:
                        solver_opts = dict(tol=tol)

                    # use current site as initial guess (iterate over site ind)
                    x0 = x_previous.get(site, b)
                    x = np.stack([
                        getattr(spla, solver)
                        (N, b[..., k], x0=x0[..., k], **solver_opts)[0]
                        for k in range(x0.shape[-1])
                    ], axis=-1)

                # update the tensors (all 'virtual' TNs above also updated)
                Tk, Tb = ket[site], bra[site]
                Tk.modify(data=reshape(x, Tk.shape))
                Tb.modify(data=reshape(conj(x), Tb.shape))

                # store solution to check convergence
                xs[site] = x

            # after updating both sites check for convergence of tensor entries
            cost_fid = do('trace', do('real', dag(x) @ b))
            cost_norm = do('abs', do('trace', dag(x) @ (N @ x)))
            cost = - 2 * cost_fid + cost_norm

            converged = (
                (previous_cost is not None) and
                (abs(cost - previous_cost) < tol)
            )
            if converged:
                break

            previous_cost = cost
            for site in tags_plq:
                x_previous[site] = xs[site]

    if condition_tensors:
        if condition_maintain_norms:
            conditioner(
                ket_plq, value=pre_norm, balance_bonds=condition_balance_bonds)
        else:
            conditioner(
                ket_plq, balance_bonds=condition_balance_bonds)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))


def gate_full_update_autodiff_fidelity(
    ket,
    env,
    bra,
    G,
    where,
    tags_plq,
    steps,
    tol,
    max_bond,
    optimize='auto-hq',
    autodiff_backend='autograd',
    autodiff_optimizer='L-BFGS-B',
    init_simple_guess=True,
    condition_tensors=True,
    condition_maintain_norms=True,
    condition_balance_bonds=True,
    **kwargs,
):
    ket_plq = ket.select_any(tags_plq).view_like_(ket)
    bra_plq = bra.select_any(tags_plq).view_like_(bra)

    # the target sites + gate and also norm (copy - not virtual)
    target = ket_plq.gate(G, where, contract=False) | env

    # make initial guess the simple gate tensors
    if init_simple_guess:
        ket_plq.gate_(G, where, contract='reduce-split', max_bond=max_bond)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))

    if condition_tensors:
        conditioner(ket_plq, balance_bonds=condition_balance_bonds)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))
        if condition_maintain_norms:
            pre_norm = ket_plq[site].norm()

    def fidelity(bra_plq):
        for site in tags_plq:
            ket_plq[site].modify(data=conj(bra_plq[site].data))

        fid = (bra_plq | target).contract(all, optimize=optimize)
        norm = (bra_plq | env | ket_plq).contract(all, optimize=optimize)

        return - 2 * do('abs', fid) + do('abs', norm)

    tnopt = TNOptimizer(
        bra_plq,
        loss_fn=fidelity,
        tags=tags_plq,
        progbar=False,
        optimizer=autodiff_optimizer,
        autodiff_backend=autodiff_backend,
        **kwargs,
    )
    bra_plq_opt = tnopt.optimize(steps, tol=tol)

    for site in tags_plq:
        new_data = bra_plq_opt[site].data
        ket[site].modify(data=conj(new_data))
        bra[site].modify(data=new_data)

    if condition_tensors:
        if condition_maintain_norms:
            conditioner(
                ket_plq, value=pre_norm, balance_bonds=condition_balance_bonds)
        else:
            conditioner(
                ket_plq, balance_bonds=condition_balance_bonds)
        for site in tags_plq:
            bra_plq[site].modify(data=conj(ket_plq[site].data))


def get_default_full_update_fit_opts():
    """The default options for the full update gate fitting procedure.
    """
    return {
        # general
        'tol': 1e-10,
        'steps': 20,
        'init_simple_guess': True,
        'condition_tensors': True,
        'condition_maintain_norms': True,
        # alternative least squares
        'als_dense': True,
        'als_solver': 'solve',
        'als_enforce_pos': False,
        'als_enforce_pos_smudge': 1e-6,
        # automatic differentation optimizing
        'autodiff_backend': 'autograd',
        'autodiff_optimizer': 'L-BFGS-B',
    }


def parse_specific_gate_opts(strategy, fit_opts):
    """Parse the options from ``fit_opts`` which are relevant for ``strategy``.
    """
    gate_opts = {
        'tol': fit_opts['tol'],
        'steps': fit_opts['steps'],
        'init_simple_guess': fit_opts['init_simple_guess'],
        'condition_tensors': fit_opts['condition_tensors'],
        'condition_maintain_norms': fit_opts['condition_maintain_norms'],
    }

    if 'als' in strategy:
        gate_opts['solver'] = fit_opts['als_solver']
        gate_opts['dense'] = fit_opts['als_dense']
        gate_opts['enforce_pos'] = fit_opts['als_enforce_pos']
        gate_opts['pos_smudge'] = fit_opts['als_enforce_pos_smudge']

    elif 'autodiff' in strategy:
        gate_opts['autodiff_backend'] = fit_opts['autodiff_backend']
        gate_opts['autodiff_optimizer'] = fit_opts['autodiff_optimizer']

    return gate_opts


class FullUpdate(TEBD2D):
    """Implements the 'Full Update' version of 2D imaginary time evolution,
    where each application of a gate is fitted to the current tensors using a
    boundary contracted environment.

    Parameters
    ----------
    psi0 : TensorNetwork2DVector
        The initial state.
    ham : LocalHam2D
        The Hamtiltonian consisting of local terms.
    tau : float, optional
        The default local exponent, if considered as time real values here
        imply imaginary time.
    max_bond : {'psi0', int, None}, optional
        The maximum bond dimension to keep when applying each gate.
    gate_opts : dict, optional
        Supplied to :meth:`quimb.tensor.tensor_2d.TensorNetwork2DVector.gate`,
        in addition to ``max_bond``. By default ``contract`` is set to
        'reduce-split' and ``cutoff`` is set to ``0.0``.
    ordering : str, tuple[tuple[int]], callable, optional
        How to order the terms, if a string is given then use this as the
        strategy given to
        :meth:`~quimb.tensor.tensor_2d_tebd.LocalHam2D.get_auto_ordering`. An
        explicit list of coordinate pairs can also be given. The default is to
        greedily form an 'edge coloring' based on the sorted list of
        Hamiltonian pair coordinates. If a callable is supplied it will be used
        to generate the ordering before each sweep.
    second_order_reflect : bool, optional
        If ``True``, then apply each layer of gates in ``ordering`` forward
        with half the time step, then the same with reverse order.
    compute_energy_every : None or int, optional
        How often to compute and record the energy. If a positive integer 'n',
        the energy is computed *before* every nth sweep (i.e. including before
        the zeroth).
    compute_energy_final : bool, optional
        Whether to compute and record the energy at the end of the sweeps
        regardless of the value of ``compute_energy_every``. If you start
        sweeping again then this final energy is the same as the zeroth of the
        next set of sweeps and won't be recomputed.
    compute_energy_opts : dict, optional
        Supplied to
        :meth:`~quimb.tensor.tensor_2d.PEPS.compute_local_expectation`. By
        default ``max_bond`` is set to ``max(8, D**2)`` where ``D`` is the
        maximum bond to use for applying the gate, ``cutoff`` is set to ``0.0``
        and ``normalized`` is set to ``True``.
    compute_energy_fn : callable, optional
        Supply your own function to compute the energy, it should take the
        ``TEBD2D`` object as its only argument.
    callback : callable, optional
        A custom callback to run after every sweep, it should take the
        ``TEBD2D`` object as its only argument. If it returns any value
        that boolean evaluates to ``True`` then terminal the evolution.
    progbar : boolean, optional
        Whether to show a live progress bar during the evolution.
    fit_strategy : {'als', 'autodiff-fidelity'}, optional
        Core method used to fit the gate application.

            * ``'als'``: alternating least squares
            * ``'autodiff-fidelity'``: local fidelity using autodiff

    fit_opts : dict, optional
        Advanced options for the gate application fitting functions. Defaults
        are inserted and can be accessed via the ``.fit_opts`` attribute.
    compute_envs_every : {'term', 'group', 'sweep', int}, optional
        How often to recompute the environments used to the fit the gate
        application:

            * ``'term'``: every gate
            * ``'group'``: every set of commuting gates (the default)
            * ``'sweep'``: every total sweep
            * int: every ``x`` number of total sweeps

    pre_normalize : bool, optional
        Actively renormalize the state using the computed environments.
    condition_tensors : bool, optional
        Whether to actively equalize tensor norms for numerical stability.
    condition_balance_bonds : bool, optional
        If and when equalizing tensor norms, whether to also balance bonds as
        an additional conditioning.
    contract_optimize : str, optional
        Contraction path optimizer to use for gate + env + sites contractions.

    Attributes
    ----------
    state : TensorNetwork2DVector
        The current state.
    ham : LocalHam2D
        The Hamiltonian being used to evolve.
    energy : float
        The current of the current state, this will trigger a computation if
        the energy at this iteration hasn't been computed yet.
    energies : list[float]
        The energies that have been computed, if any.
    its : list[int]
        The corresponding sequence of iteration numbers that energies have been
        computed at.
    taus : list[float]
        The corresponding sequence of time steps that energies have been
        computed at.
    best : dict
        If ``keep_best`` was set then the best recorded energy and the
        corresponding state that was computed - keys ``'energy'`` and
        ``'state'`` respectively.
    fit_opts : dict
        Detailed options for fitting the applied gate.
    """

    def __init__(
        self,
        psi0,
        ham,
        tau=0.01,
        D=None,
        chi=None,
        fit_strategy='als',
        fit_opts=None,
        compute_envs_every=1,
        pre_normalize=True,
        condition_tensors=True,
        condition_balance_bonds=True,
        contract_optimize='auto-hq',
        imag=True,
        gate_opts=None,
        ordering=None,
        second_order_reflect=False,
        compute_energy_every=None,
        compute_energy_final=True,
        compute_energy_opts=None,
        compute_energy_fn=None,
        compute_energy_per_site=False,
        callback=None,
        keep_best=False,
        progbar=True,
    ):
        super().__init__(
            psi0=psi0,
            ham=ham,
            tau=tau,
            D=D,
            chi=chi,
            imag=imag,
            gate_opts=gate_opts,
            ordering=ordering,
            second_order_reflect=second_order_reflect,
            compute_energy_every=compute_energy_every,
            compute_energy_final=compute_energy_final,
            compute_energy_opts=compute_energy_opts,
            compute_energy_fn=compute_energy_fn,
            compute_energy_per_site=compute_energy_per_site,
            callback=callback,
            keep_best=keep_best,
            progbar=progbar,
        )

        self.fit_strategy = str(fit_strategy)
        self.fit_opts = get_default_full_update_fit_opts()
        if fit_opts is not None:
            bad_opts = set(fit_opts) - set(self.fit_opts)
            if bad_opts:
                raise ValueError("Invalid fit option(s): {}".format(bad_opts))
            self.fit_opts.update(fit_opts)

        self.pre_normalize = bool(pre_normalize)
        self.contract_optimize = str(contract_optimize)
        self.condition_tensors = bool(condition_tensors)
        self.condition_balance_bonds = bool(condition_balance_bonds)

        self.compute_envs_every = compute_envs_every
        self._env_n = self._env_term_count = self._env_group_count = -1

        self._psi.add_tag('KET')

    @property
    def fit_strategy(self):
        return self._fit_strategy

    @fit_strategy.setter
    def fit_strategy(self, fit_strategy):
        self._gate_fit_fn = {
            'als': gate_full_update_als,
            'autodiff-fidelity': gate_full_update_autodiff_fidelity,
        }[fit_strategy]
        self._fit_strategy = fit_strategy

    def set_state(self, psi):
        self._psi = psi.copy()

        # ensure the final dimension of each tensor is the physical dim
        for tag, ind in zip(self._psi.site_tags, self._psi.site_inds):
            t = self._psi[tag]
            if t.inds[-1] != ind:
                new_inds = [i for i in t.inds if i != ind] + [ind]
                t.transpose_(*new_inds)

    @property
    def compute_envs_every(self):
        return self._compute_envs_every

    @compute_envs_every.setter
    def compute_envs_every(self, x):
        if x == 'sweep':
            self._need_to_recompute_envs = lambda: (
                (self._n != self._env_n)
            )
        elif x == 'group':
            self._need_to_recompute_envs = lambda: (
                (self._n != self._env_n) or
                (self._group_count != self._env_group_count)
            )
        elif x == 'term':
            self._need_to_recompute_envs = lambda: (
                (self._n != self._env_n) or
                (self._group_count != self._env_group_count) or
                (self._term_count != self._env_term_count)
            )
        else:
            x = max(1, int(x))
            self._need_to_recompute_envs = lambda: (self._n >= self._env_n + x)

        self._compute_envs_every = x

    def _maybe_compute_plaquette_envs(self, force=False):
        """Compute and store the plaquette environments for all local terms.
        """
        # first check if we need to compute the envs
        if not self._need_to_recompute_envs() and not force:
            return

        if self.condition_tensors:
            conditioner(self._psi, balance_bonds=self.condition_balance_bonds)

        # useful to store the bra that went into making the norm
        norm, _, self._bra = self._psi.make_norm(return_all=True)

        envs = dict()
        for x_bsz, y_bsz in calc_plaquette_sizes(self.ham.terms):
            envs.update(norm.compute_plaquette_environments(
                x_bsz=x_bsz, y_bsz=y_bsz, max_bond=self.chi, cutoff=0.0))

        if self.pre_normalize:
            # get the first plaquette env and use it to compute current norm
            p0, env0 = next(iter(envs.items()))
            sites = plaquette_to_sites(p0)
            tags_plq = tuple(starmap(norm.site_tag, sites))
            norm_plq = norm.select_any(tags_plq) | env0

            # contract the local plaquette norm
            nfactor = do(
                'abs', norm_plq.contract(all, optimize=self.contract_optimize))

            # scale the bra and ket and each of the plaquette environments
            self._psi.multiply_(nfactor**(-1 / 2), spread_over='all')
            self._bra.multiply_(nfactor**(-1 / 2), spread_over='all')

            # scale the envs, taking into account the number of sites missing
            n = self._psi.num_tensors
            for ((_, _), (di, dj)), env in envs.items():
                n_missing = di * dj
                env.multiply_(nfactor ** (n_missing / n - 1),
                              spread_over='all')

        self.plaquette_envs = envs
        self.plaquette_mapping = calc_plaquette_map(envs)

        self._env_n = self._n
        self._env_group_count = self._group_count
        self._env_term_count = self._term_count

    def presweep(self, i):
        """Full update presweep - compute envs and inject gate options.
        """
        # inject the specific gate options required (do
        # here so user can change options between sweeps)
        self._gate_opts = parse_specific_gate_opts(
            self.fit_strategy, self.fit_opts)

        # keep track of number of gates applied, and commutative groups
        self._term_count = 0
        self._group_count = 0
        self._current_group = set()

    def compute_energy(self):
        """Full update compute energy - use the (likely) already calculated
        plaquette environments.
        """
        self._maybe_compute_plaquette_envs(force=self._n != self._env_n)

        return self.state.compute_local_expectation(
            self.ham.terms,
            plaquette_envs=self.plaquette_envs,
            plaquette_mapping=self.plaquette_mapping,
            **self.compute_energy_opts
        )

    def gate(self, G, where):
        """Apply the gate ``G`` at sites where, using a fitting method that
        takes into account the current environment.
        """
        # check if the new term commutes with those applied so far, this is to
        #     decide if we need to recompute the environments
        swhere = set(where)
        if self._current_group.isdisjoint(swhere):
            # if so add it to the grouping
            self._current_group |= swhere
        else:
            # else increment and reset the grouping
            self._current_group = swhere
            self._group_count += 1

        # get the plaquette containing ``where`` and the sites it contains -
        # these will all be fitted
        self._maybe_compute_plaquette_envs()
        plq = self.plaquette_mapping[tuple(sorted(where))]
        env = self.plaquette_envs[plq]
        tags_plq = tuple(starmap(self._psi.site_tag, plaquette_to_sites(plq)))

        # perform the gate, inplace
        self._gate_fit_fn(
            ket=self._psi,
            env=env,
            bra=self._bra,
            G=G,
            where=where,
            tags_plq=tags_plq,
            max_bond=self.D,
            optimize=self.contract_optimize,
            condition_balance_bonds=self.condition_balance_bonds,
            **self._gate_opts
        )

        # increments every gate call regardless
        self._term_count += 1



import pytest

import numpy as np
from numpy.testing import assert_allclose

import quimb as qu
from quimb.tensor import (
    MatrixProductState, MatrixProductOperator, tensor_network_align,
    MPS_rand_state, MPO_identity, MPO_identity_like, MPO_zeros, MPO_zeros_like,
    MPO_rand, MPO_rand_herm, MPO_ham_heis, MPS_neel_state, MPS_zero_state,
    bonds, MPS_computational_state, Dense1D)
from quimb.tensor.tensor_core import oset


dtypes = ['float32', 'float64', 'complex64', 'complex128']


class TestMatrixProductState:
    def test_matrix_product_state(self):
        tensors = ([np.random.rand(5, 2)] +
                   [np.random.rand(5, 5, 2)
                    for _ in range(3)] + [np.random.rand(5, 2)])
        mps = MatrixProductState(tensors)
        assert len(mps.tensors) == 5
        nmps = mps.reindex_sites('foo{}', inplace=False, where=slice(0, 3))
        assert nmps.site_ind_id == "k{}"
        assert isinstance(nmps, MatrixProductState)
        assert set(nmps.outer_inds()) == {'foo0', 'foo1', 'foo2', 'k3', 'k4'}
        assert set(mps.outer_inds()) == {'k0', 'k1', 'k2', 'k3', 'k4'}
        mps.site_ind_id = 'foo{}'
        assert set(
            mps.outer_inds()) == {'foo0', 'foo1', 'foo2', 'foo3', 'foo4'}
        assert mps.site_inds == ('foo0', 'foo1', 'foo2', 'foo3', 'foo4')
        assert mps.site_ind_id == 'foo{}'
        mps.show()

    @pytest.mark.parametrize(
        "dtype", [float, complex, np.complex128, np.float64, 'raise'])
    def test_rand_mps_dtype(self, dtype):
        if dtype == 'raise':
            with pytest.raises(TypeError):
                MPS_rand_state(10, 7, dtype=dtype)
        else:
            p = MPS_rand_state(10, 7, dtype=dtype)
            assert p[0].dtype == dtype
            assert p[7].dtype == dtype

    def test_trans_invar(self):
        with pytest.raises(ValueError):
            psi = MPS_rand_state(10, 7, cyclic=False, trans_invar=True)

        psi = MPS_rand_state(10, 7, cyclic=True, trans_invar=True)
        z0 = psi.expec(psi.gate(qu.pauli('Z'), 0, contract=True))
        z3 = psi.expec(psi.gate(qu.pauli('Z'), 0, contract=True))
        z7 = psi.expec(psi.gate(qu.pauli('Z'), 0, contract=True))

        assert_allclose(z0, z3)
        assert_allclose(z3, z7)

    def test_from_dense(self):
        psi = qu.rand_ket(2**8)
        mps = MatrixProductState.from_dense(psi, dims=[2] * 8)
        assert mps.tags == oset(f'I{i}' for i in range(8))
        assert mps.site_inds == tuple(f'k{i}' for i in range(8))
        assert mps.L == 8
        mpod = mps.to_dense()
        assert qu.expec(mpod, psi) == pytest.approx(1)

    def test_left_canonize_site(self):
        a = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        b = np.random.randn(7, 7, 2) + 1.0j * np.random.randn(7, 7, 2)
        c = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        mps = MatrixProductState([a, b, c], site_tag_id="I{}")

        mps.left_canonize_site(0)
        assert mps['I0'].shape == (2, 2)
        assert mps['I0'].tags == oset(('I0',))
        assert mps['I1'].tags == oset(('I1',))

        U = (mps['I0'].data)
        assert_allclose(U.conj().T @ U, np.eye(2), atol=1e-13)
        assert_allclose(U @ U.conj().T, np.eye(2), atol=1e-13)

        # combined two site contraction is identity also
        mps.left_canonize_site(1)
        ptn = (mps.H & mps) ^ ['I0', 'I1']
        assert_allclose(ptn['I1'].data, np.eye(4), atol=1e-13)

        # try normalizing the state
        mps['I2'] /= mps['I2'].norm()

        assert_allclose(abs(mps.H @ mps), 1.0)

    def test_right_canonize_site(self):
        a = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        b = np.random.randn(7, 7, 2) + 1.0j * np.random.randn(7, 7, 2)
        c = np.random.randn(7, 2) + 1.0j * np.random.randn(7, 2)
        mps = MatrixProductState([a, b, c], site_tag_id="I{}")

        mps.right_canonize_site(2)
        assert mps['I2'].shape == (2, 2)
        assert mps['I2'].tags == oset(('I2',))
        assert mps['I1'].tags == oset(('I1',))

        U = (mps['I2'].data)
        assert_allclose(U.conj().T @ U, np.eye(2), atol=1e-13)
        assert_allclose(U @ U.conj().T, np.eye(2), atol=1e-13)

        # combined two site contraction is identity also
        mps.right_canonize_site(1)
        ptn = (mps.H & mps) ^ ['I1', 'I2']
        assert_allclose(ptn['I1'].data, np.eye(4), atol=1e-13)

        # try normalizing the state
        mps['I0'] /= mps['I0'].norm()

        assert_allclose(mps.H @ mps, 1)

    def test_rand_mps_left_canonize(self):
        n = 10
        k = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', normalize=False)
        k.left_canonize(normalize=True)

        assert k.count_canonized() == (9, 0)

        assert_allclose(k.H @ k, 1)
        p_tn = (k.H & k) ^ slice(0, 9)
        assert_allclose(p_tn['foo8'].data, np.eye(10), atol=1e-13)

    def test_rand_mps_left_canonize_with_bra(self):
        n = 10
        k = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', normalize=False)
        b = k.H
        k.left_canonize(normalize=True, bra=b)
        assert_allclose(b @ k, 1)
        p_tn = (b & k) ^ slice(0, 9)
        assert_allclose(p_tn['foo8'].data, np.eye(10), atol=1e-13)

    def test_rand_mps_right_canonize(self):
        n = 10
        k = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', normalize=False)
        k.right_canonize(normalize=True)
        assert_allclose(k.H @ k, 1)
        p_tn = (k.H & k) ^ slice(..., 0, -1)
        assert_allclose(p_tn['foo1'].data, np.eye(10), atol=1e-13)

    def test_rand_mps_right_canonize_with_bra(self):
        n = 10
        k = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', normalize=False)
        b = k.H
        k.right_canonize(normalize=True, bra=b)
        assert_allclose(b @ k, 1)
        p_tn = (b & k) ^ slice(..., 0, -1)
        assert_allclose(p_tn['foo1'].data, np.eye(10), atol=1e-13)

    def test_rand_mps_mixed_canonize(self):
        n = 10
        rmps = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', normalize=True)

        # move to the center
        rmps.canonize(4)
        assert rmps.count_canonized() == (4, 5)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 4) ^ slice(..., 4, -1)
        assert_allclose(p_tn['foo3'].data, np.eye(10), atol=1e-13)
        assert_allclose(p_tn['foo5'].data, np.eye(10), atol=1e-13)

        # try shifting to the right
        rmps.shift_orthogonality_center(current=4, new=8)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 8) ^ slice(..., 8, -1)
        assert_allclose(p_tn['foo7'].data, np.eye(4), atol=1e-13)
        assert_allclose(p_tn['foo9'].data, np.eye(2), atol=1e-13)

        # try shifting to the left
        rmps.shift_orthogonality_center(current=8, new=6)
        assert_allclose(rmps.H @ rmps, 1)
        p_tn = (rmps.H & rmps) ^ slice(0, 6) ^ slice(..., 6, -1)
        assert_allclose(p_tn['foo5'].data, np.eye(10), atol=1e-13)
        assert_allclose(p_tn['foo7'].data, np.eye(8), atol=1e-13)

    @pytest.mark.parametrize("dtype", dtypes)
    def test_canonize_and_calc_current_orthog_center(self, dtype):
        p = MPS_rand_state(20, 3, dtype=dtype)
        co = p.calc_current_orthog_center()
        assert co == (0, 19)
        p.canonize((5, 15), co)
        co = p.calc_current_orthog_center()
        assert co == (5, 15)
        p.canonize((8, 11), co)
        co = p.calc_current_orthog_center()
        assert co == (8, 11)
        assert p.dtype == dtype

    def test_can_change_data(self):
        p = MPS_rand_state(3, 10)
        assert_allclose(p.H @ p, 1)
        p[1].modify(data=np.random.randn(10, 10, 2))
        assert abs(p.H @ p - 1) > 1e-13

    def test_can_change_data_using_subnetwork(self):
        p = MPS_rand_state(3, 10)
        pH = p.H
        p.add_tag('__ket__')
        pH.add_tag('__bra__')
        tn = p | pH
        assert_allclose((tn ^ ...), 1)
        assert_allclose(tn[('__ket__', 'I1')].data, tn[('__bra__',
                                                        'I1')].data.conj())
        p[1].modify(data=np.random.randn(10, 10, 2))
        assert abs((tn ^ ...) - 1) > 1e-13
        assert not np.allclose(tn[('__ket__', 'I1')].data,
                               tn[('__bra__', 'I1')].data.conj())

    def test_adding_mps(self):
        p = MPS_rand_state(10, 7)
        assert max(p['I4'].shape) == 7
        p2 = p + p
        assert max(p2['I4'].shape) == 14
        assert_allclose(p2.H @ p, 2)
        p += p
        assert max(p['I4'].shape) == 14
        assert_allclose(p.H @ p, 4)

    @pytest.mark.parametrize("method", ['svd', 'eig'])
    @pytest.mark.parametrize('cutoff_mode', ['abs', 'rel', 'sum2'])
    def test_compress_mps(self, method, cutoff_mode):
        n = 10
        chi = 7
        p = MPS_rand_state(n, chi)
        assert max(p['I4'].shape) == chi
        p2 = p + p
        assert max(p2['I4'].shape) == chi * 2
        assert_allclose(p2.H @ p, 2)
        p2.left_compress(method=method, cutoff=1e-6, cutoff_mode=cutoff_mode)
        assert max(p2['I4'].shape) == chi
        assert_allclose(p2.H @ p, 2)
        assert p2.count_canonized() == (n - 1, 0)

    def test_compress_mps_right(self):
        p = MPS_rand_state(10, 7)
        assert max(p['I4'].shape) == 7
        p2 = p + p
        assert max(p2['I4'].shape) == 14
        assert_allclose(p2.H @ p, 2)
        p2.right_compress()
        assert max(p2['I4'].shape) == 7
        assert_allclose(p2.H @ p, 2)

    @pytest.mark.parametrize("method", ['svd', 'eig'])
    def test_compress_trim_max_bond(self, method):
        p0 = MPS_rand_state(20, 20)
        p = p0.copy()
        p.compress(method=method)
        assert max(p['I4'].shape) == 20
        p.compress(max_bond=13, method=method)
        assert max(p['I4'].shape) == 13
        assert_allclose(p.H @ p, p0.H @ p0)

    def test_compress_form(self):
        p = MPS_rand_state(20, 20)
        p.compress('left')
        assert p.count_canonized() == (19, 0)
        p.compress('right')
        assert p.count_canonized() == (0, 19)
        p.compress(7)
        assert p.count_canonized() == (7, 12)
        p = MPS_rand_state(20, 20)
        p.compress('flat', absorb='left')
        assert p.count_canonized() == (0, 0)

    def test_compress_site(self):
        psi = MPS_rand_state(10, 7)
        psi.compress_site(3, max_bond=1)
        assert psi.bond_sizes() == [2, 4, 1, 1, 7, 7, 7, 4, 2]
        assert psi.calc_current_orthog_center() == (3, 3)

        psi = MPS_rand_state(10, 7)
        psi.compress_site(0, max_bond=1)
        assert psi.bond_sizes() == [1, 7, 7, 7, 7, 7, 7, 4, 2]
        assert psi.calc_current_orthog_center() == (0, 0)

        psi = MPS_rand_state(10, 7)
        psi.compress_site(9, max_bond=1)
        assert psi.bond_sizes() == [2, 4, 7, 7, 7, 7, 7, 7, 1]
        assert psi.calc_current_orthog_center() == (9, 9)

    @pytest.mark.parametrize("method", ['svd', 'eig'])
    @pytest.mark.parametrize("form", ['left', 'right', 'raise'])
    def test_add_and_compress_mps(self, method, form):
        p = MPS_rand_state(10, 7)
        assert max(p['I4'].shape) == 7

        if form == 'raise':
            with pytest.raises(ValueError):
                p.add_MPS(
                    p, compress=True, method=method, form=form, cutoff=1e-6)
            return

        p2 = p.add_MPS(p, compress=True, method=method, form=form, cutoff=1e-6)
        assert max(p2['I4'].shape) == 7
        assert_allclose(p2.H @ p, 2)

    def test_subtract(self):
        a, b, c = (MPS_rand_state(10, 7) for _ in 'abc')
        ab = a.H @ b
        ac = a.H @ c
        abmc = a.H @ (b - c)
        assert_allclose(ab - ac, abmc)

    def test_subtract_inplace(self):
        a, b, c = (MPS_rand_state(10, 7) for _ in 'abc')
        ab = a.H @ b
        ac = a.H @ c
        b -= c
        abmc = a.H @ b
        assert_allclose(ab - ac, abmc)

    def test_amplitude(self):
        mps = MPS_rand_state(10, 7)
        k = mps.to_dense()
        idx = np.random.randint(0, k.shape[0])
        c_b = mps.amplitude(f'{idx:0>10b}')
        assert k[idx, 0] == pytest.approx(c_b)

    def test_schmidt_values_entropy_gap_simple(self):
        n = 12
        p = MPS_rand_state(n, 16)
        p.right_canonize()
        svns = []
        sgs = []
        for i in range(1, n):
            sgs.append(p.schmidt_gap(i, cur_orthog=i - 1))
            svns.append(p.entropy(i, cur_orthog=i))

        pd = p.to_dense()
        ex_svns = [
            qu.entropy_subsys(pd, [2] * n, range(i)) for i in range(1, n)
        ]
        ex_sgs = [qu.schmidt_gap(pd, [2] * n, range(i)) for i in range(1, n)]
        assert_allclose(ex_svns, svns)
        assert_allclose(ex_sgs, sgs)

    def test_magnetization(self):
        binary = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
        p = MPS_computational_state(binary)
        mzs = [p.magnetization(i) for i in range(len(binary))]
        assert_allclose(mzs, 0.5 - np.array(binary))

    @pytest.mark.parametrize("rescale", [False, True])
    @pytest.mark.parametrize(
        "keep", [(2, 3, 4, 6, 8), slice(-2, 4), slice(3, -1, -1), [1]])
    def test_partial_trace(self, rescale, keep):
        n = 10
        p = MPS_rand_state(n, 7)
        r = p.ptr(keep=keep, upper_ind_id='u{}', rescale_sites=rescale)
        rd = r.to_dense()
        if isinstance(keep, slice):
            keep = p.slice2sites(keep)
        else:
            if rescale:
                if keep == [1]:
                    assert r.lower_inds_present == ('u0',)
                    assert r.upper_inds_present == ('k0',)
                else:
                    assert r.lower_inds_present == ('u0', 'u1', 'u2',
                                                   'u3', 'u4')
                    assert r.upper_inds_present == ('k0', 'k1', 'k2',
                                                   'k3', 'k4')
            else:
                if keep == [1]:
                    assert r.lower_inds_present == ('u1',)
                    assert r.upper_inds_present == ('k1',)
                else:
                    assert r.lower_inds_present == ('u2', 'u3', 'u4',
                                                    'u6', 'u8')
                    assert r.upper_inds_present == ('k2', 'k3', 'k4',
                                                    'k6', 'k8')
        assert_allclose(r.trace(), 1.0)
        assert qu.isherm(rd)
        pd = p.to_dense()
        rdd = pd.ptr([2] * n, keep=keep)
        assert_allclose(rd, rdd)

    def test_bipartite_schmidt_state(self):
        psi = MPS_rand_state(16, 5)
        psid = psi.to_dense()
        eln = qu.logneg(psid, [2**7, 2**9])

        s_d_ket = psi.bipartite_schmidt_state(7, get='ket-dense')
        ln_d_ket = qu.logneg(s_d_ket, [5, 5])
        assert_allclose(eln, ln_d_ket, rtol=1e-5)

        s_d_rho = psi.bipartite_schmidt_state(7, get='rho-dense')
        ln_d_rho = qu.logneg(s_d_rho, [5, 5])
        assert_allclose(eln, ln_d_rho, rtol=1e-5)

        T_s_ket = psi.bipartite_schmidt_state(7, get='ket')
        assert set(T_s_ket.inds) == {'kA', 'kB'}
        assert_allclose(T_s_ket.H @ T_s_ket, 1.0)

        T_s_rho = psi.bipartite_schmidt_state(7, get='rho')
        assert set(T_s_rho.outer_inds()) == {'kA', 'kB', 'bA', 'bB'}
        assert_allclose(T_s_rho.H @ T_s_rho, 1.0)

    @pytest.mark.parametrize(
        "method", ['isvd', 'svds', ('isvd', 'eigsh'), ('isvd', 'cholesky')])
    @pytest.mark.parametrize("cyclic", [True, False])
    @pytest.mark.parametrize(
        "sysa", [range(0, 10),
                 range(10, 20),
                 range(20, 30),
                 range(0, 30)])
    @pytest.mark.parametrize(
        "sysb", [range(30, 40),
                 range(40, 50),
                 range(50, 60),
                 range(30, 60)])
    def test_partial_trace_compress(self, method, cyclic, sysa, sysb):
        k = MPS_rand_state(60, 5, cyclic=cyclic)
        kws = dict(sysa=sysa, sysb=sysb, eps=1e-6, method=method, verbosity=2)
        rhoc_ab = k.partial_trace_compress(**kws)
        assert set(rhoc_ab.outer_inds()) == {'kA', 'kB', 'bA', 'bB'}
        inds = ['kA', 'kB'], ['bA', 'bB']
        x = rhoc_ab.trace(*inds)
        assert_allclose(1.0, x, rtol=1e-3)

    @pytest.mark.parametrize("cyclic", [True, False])
    def test_known_bad_case(self, cyclic):
        k = MPS_rand_state(5, 10, cyclic=cyclic)
        rhoc_ab = k.partial_trace_compress(sysa=range(2), sysb=range(2, 4))
        inds = ['kA', 'kB'], ['bA', 'bB']
        x = rhoc_ab.trace(*inds)
        assert_allclose(1.0, x, rtol=1e-3)

    @pytest.mark.parametrize("block", [
        0, 20, 39,
        slice(0, 5),
        slice(20, 25),
        slice(35, 40),
        slice(38, 42),
        slice(-3, 2)
    ])
    @pytest.mark.parametrize("dtype", [float, complex])
    def test_canonize_cyclic(self, dtype, block):
        k = MPS_rand_state(40, 10, dtype=dtype, cyclic=True)
        b = k.H
        k.add_tag('KET')
        b.add_tag('BRA')
        kb = (b | k)

        assert not np.allclose(k[block].H @ k[block], 1.0)
        assert not np.allclose(b[block].H @ b[block], 1.0)
        k.canonize_cyclic(block, bra=b)
        assert_allclose(k[block].H @ k[block], 1.0, rtol=2e-4)
        assert_allclose(b[block].H @ b[block], 1.0, rtol=2e-4)

        ii = kb.select(block, which='!any') ^ all

        if isinstance(block, slice):
            start, stop = block.start, block.stop
        else:
            start, stop = block, block + 1

        assert len(kb.select_tensors(block, 'any')) == 2 * (stop - start)

        ul, = bonds(kb[k.site_tag(start - 1), 'BRA'],
                    kb[k.site_tag(start), 'BRA'])
        ur, = bonds(kb[k.site_tag(stop - 1), 'BRA'],
                    kb[k.site_tag(stop), 'BRA'])
        ll, = bonds(kb[k.site_tag(start - 1), 'KET'],
                    kb[k.site_tag(start), 'KET'])
        lr, = bonds(kb[k.site_tag(stop - 1), 'KET'],
                    kb[k.site_tag(stop), 'KET'])

        ii = ii.to_dense((ul, ur), (ll, lr))
        assert_allclose(ii, np.eye(ii.shape[0]), rtol=0.001, atol=0.001)

    @pytest.mark.parametrize("bsz", [1, 2])
    @pytest.mark.parametrize("propagate_tags", [False, True])
    @pytest.mark.parametrize("contract", [False, True])
    def test_gate_no_contract(self, bsz, propagate_tags, contract):
        p = MPS_rand_state(5, 7, tags={'PSI0'})
        q = p.copy()
        G = qu.rand_uni(2**bsz)
        p = p.gate_(G, where=[i for i in range(2, 2 + bsz)], tags='G',
                    contract=contract, propagate_tags=propagate_tags)
        TG = p['G']
        if propagate_tags or contract:
            assert p.site_tag(2) in TG.tags
        assert ('PSI0' in TG.tags) == (propagate_tags is True) or contract
        assert (p.H & p) ^ all == pytest.approx(1.0)
        assert abs((q.H & p) ^ all) < 1.0
        assert len(p.tensors) == 6 - int(contract) * bsz
        assert set(p.outer_inds()) == {f'k{i}' for i in range(5)}

    @pytest.mark.parametrize("propagate_tags", [False, 'sites',
                                                'register', True])
    def test_gate_split_gate(self, propagate_tags):
        p = MPS_rand_state(5, 7, tags={'PSI0'})
        q = p.copy()
        G = qu.CNOT()
        p = p.gate_(G, where=[i for i in range(2, 4)], tags='G',
                    contract='split-gate', propagate_tags=propagate_tags)

        TG = sorted(p['G'], key=lambda t: sorted(t.tags))

        if propagate_tags is False:
            assert TG[0].tags == oset(('G',))
            assert TG[1].tags == oset(('G',))

        elif propagate_tags == 'register':
            assert TG[0].tags == oset(['G', 'I2'])
            assert TG[1].tags == oset(['G', 'I3'])

        elif propagate_tags == 'sites':
            assert TG[0].tags == oset(['G', 'I2', 'I3'])
            assert TG[1].tags == oset(['G', 'I2', 'I3'])

        elif propagate_tags is True:
            assert TG[0].tags == oset(['PSI0', 'G', 'I2', 'I3'])
            assert TG[1].tags == oset(['PSI0', 'G', 'I2', 'I3'])

        assert (p.H & p) ^ all == pytest.approx(1.0)
        assert abs((q.H & p) ^ all) < 1.0
        assert len(p.tensors) == 7
        assert set(p.outer_inds()) == {f'k{i}' for i in range(5)}

    def test_gate_swap_and_split(self):
        n = 10
        p = MPS_computational_state('0' * n)
        assert p.bond_sizes() == [1] * (n - 1)
        G = qu.rand_uni(4)
        p.gate_(G, (1, n - 2), contract='swap+split')
        assert p.bond_sizes() == [1] + [2] * (n - 3) + [1]

    def test_correlation(self):
        ghz = (MPS_computational_state('0000') +
               MPS_computational_state('1111')) / 2**0.5

        assert ghz.correlation(qu.pauli('Z'), 0, 1) == pytest.approx(1.0)
        assert ghz.correlation(qu.pauli('Z'), 1, 2) == pytest.approx(1.0)
        assert ghz.correlation(qu.pauli('Z'), 3, 1) == pytest.approx(1.0)
        assert ghz.correlation(
            qu.pauli('Z'), 3, 1, B=qu.pauli('Y')) == pytest.approx(0.0)

        assert ghz.H @ ghz == pytest.approx(1.0)

    def test_gate_split(self):
        psi = MPS_rand_state(10, 3)
        psi2 = psi.copy()
        G = qu.eye(2) & qu.eye(2)
        psi.gate_split_(G, (2, 3), cutoff=0)
        assert psi.bond_size(2, 3) == 6
        assert psi.H @ psi2 == pytest.approx(1.0)

        # check a unitary application
        G = qu.rand_uni(2**2)
        psi.gate_split_(G, (7, 8))
        psi.compress()
        assert psi.bond_size(2, 3) == 3
        assert psi.bond_size(7, 8) > 3
        assert psi.H @ psi == pytest.approx(1.0)
        assert abs(psi2.H @ psi) < 1.0

        # check matches dense application of gate
        psid = psi2.to_dense()
        Gd = qu.ikron(G, [2] * 10, (7, 8))
        assert psi.to_dense().H @ (Gd @ psid) == pytest.approx(1.0)

    def test_swap_far_sites(self):
        psi = MPS_rand_state(7, 2)
        for i, j in [(0, 6), (6, 1), (5, 2)]:
            k1 = psi.to_dense([
                psi.site_ind(
                    j if site == i else
                    i if site == j else
                    site
                ) for site in psi.sites
            ])
            k2 = psi.swap_sites_with_compress(i, j).to_dense()
            assert qu.fidelity(k1, k2) == pytest.approx(1.0)

    def test_swap_gating(self):
        psi0 = MPS_rand_state(20, 5)
        CNOT = qu.controlled('not')
        psi0XX = psi0.gate(CNOT, (4, 13))
        psi0XX_s = psi0.gate_with_auto_swap(CNOT, (4, 13))
        assert psi0XX.H @ psi0XX_s == pytest.approx(1.0)

    def test_auto_split_detection(self):
        psi0 = MPS_computational_state('00')
        CNOT = qu.controlled('not')
        ISWAP = qu.iswap()
        G = qu.rand_uni(4)

        opts = {'contract': 'auto-split-gate', 'where': (0, 1)}

        psi_cnot = psi0.gate(CNOT, **opts)
        psi_iswap = psi0.gate(ISWAP, **opts)
        psi_G = psi0.gate(G, **opts)

        assert (psi_cnot.max_bond() ==
                psi_iswap.max_bond() ==
                psi_G.max_bond() == 2)

        assert len(psi_cnot.tensors) == len(psi_iswap.tensors) == 4
        assert len(psi_G.tensors) == 3

    @pytest.mark.parametrize('cur_orthog', (None, 3))
    @pytest.mark.parametrize('site', (0, 5, 9))
    @pytest.mark.parametrize('outcome', (None, 2))
    @pytest.mark.parametrize('renorm', (True, False))
    @pytest.mark.parametrize('remove', (True, False))
    def test_mps_measure(self, cur_orthog, site, outcome, renorm, remove):
        psi = MPS_rand_state(10, 7, phys_dim=3, dtype=complex)
        if cur_orthog:
            psi.canonize(cur_orthog)
        outcome, psim = psi.measure(
            site, outcome=outcome, cur_orthog=cur_orthog,
            renorm=renorm, remove=remove)
        newL = 10 - int(remove)
        assert psim.L == newL
        assert psim.num_tensors == newL
        assert set(psim.site_tags) == {f'I{i}' for i in range(newL)}
        assert set(psim.site_inds) == {f'k{i}' for i in range(newL)}
        if renorm:
            assert psim.H @ psim == pytest.approx(1.0)
        else:
            assert 0.0 < psim.H @ psim < 1.0
        new_can_cen = min(site, newL - 1)
        t = psim[new_can_cen]
        if renorm:
            assert t.H @ t == pytest.approx(1.0)
        else:
            0.0 < t.H @ t < 1.0

    def test_measure_known_outcome(self):
        mps = MPS_computational_state('010101')
        assert mps.measure_(3, get='outcome') == 1

    def test_permute_arrays(self):
        mps = MPS_rand_state(7, 5)
        k0 = mps.to_dense()
        mps.canonize(3)
        mps.permute_arrays('prl')
        assert mps[0].shape == (2, 2)
        assert mps[1].shape == (2, 4, 2)
        assert mps[2].shape == (2, 5, 4)
        kf = mps.to_dense()
        assert qu.fidelity(k0, kf) == pytest.approx(1.0)


class TestMatrixProductOperator:

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_matrix_product_operator(self, cyclic):

        end_shape = (5, 5, 2, 2) if cyclic else (5, 2, 2)

        tensors = ([np.random.rand(*end_shape)] +
                   [np.random.rand(5, 5, 2, 2)
                    for _ in range(3)] + [np.random.rand(*end_shape)])
        mpo = MatrixProductOperator(tensors)

        mpo.show()
        assert len(mpo.tensors) == 5
        assert mpo.upper_inds == ('k0', 'k1', 'k2', 'k3', 'k4')
        assert mpo.lower_inds == ('b0', 'b1', 'b2', 'b3', 'b4')
        op = mpo ^ ...
        # this would rely on left to right contraction if not in set form
        assert set(op.inds) == {
            'k0', 'b0', 'k1', 'b1', 'k2', 'b2', 'k3', 'b3', 'k4', 'b4'
        }

        assert set(mpo.site_tags) == {f'I{i}' for i in range(5)}
        assert all(f'I{i}' in mpo.tags for i in range(5))
        mpo.site_tag_id = 'TEST1,{}'
        assert set(mpo.site_tags) == {f'TEST1,{i}' for i in range(5)}
        assert not any(f'I{i}' in mpo.tags for i in range(5))
        assert all(f'TEST1,{i}' in mpo.tags for i in range(5))

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_compress_mpo(self, cyclic):
        A = MPO_rand(12, 5, cyclic=cyclic)
        assert all(b == 5 for b in A.bond_sizes())
        A.expand_bond_dimension(10)
        assert all(b == 10 for b in A.bond_sizes())
        A.compress()
        assert all(b in (4, 5) for b in A.bond_sizes())

    def test_add_mpo(self):
        h = MPO_rand_herm(12, 5)
        h2 = h + h
        assert max(h2[6].shape) == 10
        t = h.trace()
        t2 = h2.trace()
        assert_allclose(2 * t, t2)

    def test_adding_mpo(self):
        h = MPO_ham_heis(6)
        hd = h.to_dense()
        assert_allclose(h @ h.H, (hd @ hd.H).tr())
        h2 = h + h
        assert_allclose(h2 @ h2.H, (hd @ hd.H).tr() * 4)
        h2.right_compress()
        assert_allclose(h2 @ h2.H, (hd @ hd.H).tr() * 4)
        assert max(h2['I3'].shape) == 5

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_subtract_mpo(self, cyclic):
        a, b = MPO_rand(13, 7, cyclic=cyclic), MPO_rand(13, 7, cyclic=cyclic)
        x1 = a.trace() - b.trace()
        assert_allclose(x1, (a - b).trace())
        a -= b
        assert_allclose(x1, a.trace())

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("rand_strength", (0, 1e-9))
    def test_expand_mpo(self, cyclic, rand_strength):
        h = MPO_ham_heis(12, cyclic=cyclic)
        assert h[0].dtype == float
        he = h.expand_bond_dimension(13, rand_strength=rand_strength)
        assert h[0].dtype == float
        assert max(he[6].shape) == 13

        if cyclic:
            assert he.bond_size(0, -1) == 13

        t = h.trace()
        te = he.trace()
        assert_allclose(t, te)

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("rand_strength", (0, 1e-9))
    def test_expand_mpo_limited(self, cyclic, rand_strength):
        h = MPO_ham_heis(12, cyclic=cyclic)
        he = h.expand_bond_dimension(3, rand_strength=rand_strength)
        # should do nothing
        assert max(he[6].shape) == 5

    def test_mpo_identity(self):
        k = MPS_rand_state(13, 7)
        b = MPS_rand_state(13, 7)
        o1 = k @ b
        i = MPO_identity(13)
        k, i, b = tensor_network_align(k, i, b)
        o2 = (k & i & b) ^ ...
        assert_allclose(o1, o2)

    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("dtype", (complex, float))
    def test_mpo_rand_herm_and_trace(self, dtype, cyclic):
        op = MPO_rand_herm(
            20, bond_dim=5, phys_dim=3, dtype=dtype, cyclic=cyclic)
        assert_allclose(op.H @ op, 1.0)
        tr_val = op.trace()
        assert tr_val != 0.0
        assert_allclose(tr_val.imag, 0.0, atol=1e-14)

    @pytest.mark.parametrize("cyclic", [False, True])
    def test_mpo_rand_herm_trace_and_identity_like(self, cyclic):
        op = MPO_rand_herm(
            20, bond_dim=5, phys_dim=3, upper_ind_id='foo{}', cyclic=cyclic)
        t = op.trace()
        assert t != 0.0
        Id = MPO_identity_like(op)
        assert_allclose(Id.trace(), 3**20)
        Id[0] *= 3 / 3**20
        op += Id
        assert_allclose(op.trace(), t + 3)

    def test_partial_transpose(self):
        p = MPS_rand_state(8, 10)
        r = p.ptr([2, 3, 4, 5, 6, 7])
        rd = r.to_dense()

        assert qu.isherm(rd)
        assert qu.ispos(rd)

        rpt = r.partial_transpose([0, 1, 2])
        rptd = rpt.to_dense()

        upper_inds = tuple(f'b{i}' for i in range(6))
        lower_inds = tuple(f'k{i}' for i in range(6))
        outer_inds = rpt.outer_inds()
        assert all(i in outer_inds for i in upper_inds + lower_inds)

        assert qu.isherm(rptd)
        assert not qu.ispos(rptd)

    def test_upper_lower_ind_id_guard(self):
        A = MPO_rand(8, 5)
        with pytest.raises(ValueError):
            A.upper_ind_id = 'b{}'
        with pytest.raises(ValueError):
            A.lower_ind_id = 'k{}'

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_apply_mpo(self, cyclic):
        A = MPO_rand(8, 5, cyclic=cyclic)
        B = MPO_rand(
            8, 5, upper_ind_id='q{}', lower_ind_id='w{}', cyclic=cyclic)
        C = A.apply(B)
        assert C.max_bond() == 25
        assert C.upper_ind_id == 'q{}'
        assert C.lower_ind_id == 'w{}'
        Ad, Bd, Cd = A.to_dense(), B.to_dense(), C.to_dense()
        assert_allclose(Ad @ Bd, Cd)

    @pytest.mark.parametrize("cyclic", (False, True))
    @pytest.mark.parametrize("site_ind_id", ('k{}', 'test{}'))
    def test_apply_mps(self, cyclic, site_ind_id):
        A = MPO_rand(8, 5, cyclic=cyclic)
        x = MPS_rand_state(8, 4, site_ind_id=site_ind_id, cyclic=cyclic)
        y = A.apply(x)
        assert y.max_bond() == 20
        assert isinstance(y, MatrixProductState)
        assert len(y.tensors) == 8
        assert y.site_ind_id == site_ind_id
        Ad, xd, yd = A.to_dense(), x.to_dense(), y.to_dense()
        assert_allclose(Ad @ xd, yd)

    def test_permute_arrays(self):
        mpo = MPO_rand(4, 3)
        A0 = mpo.to_dense()
        mpo.permute_arrays('drul')
        assert mpo[0].shape == (2, 3, 2)
        assert mpo[1].shape == (2, 3, 2, 3)
        Af = mpo.to_dense()
        assert_allclose(A0, Af)

# --------------------------------------------------------------------------- #
#                         Test specific 1D instances                          #
# --------------------------------------------------------------------------- #


class TestSpecificStatesOperators:
    @pytest.mark.parametrize("cyclic", [False, True])
    def test_rand_ket_mps(self, cyclic):
        n = 10
        rmps = MPS_rand_state(
            n, 10, site_tag_id="foo{}", tags='bar', cyclic=cyclic)
        assert rmps[0].tags == oset(['foo0', 'bar'])
        assert rmps[3].tags == oset(['foo3', 'bar'])
        assert rmps[-1].tags == oset(['foo9', 'bar'])

        rmpsH_rmps = rmps.H & rmps
        assert len(rmpsH_rmps.tag_map['foo0']) == 2
        assert len(rmpsH_rmps.tag_map['bar']) == n * 2

        assert_allclose(rmps.H @ rmps, 1)
        c = (rmps.H & rmps) ^ slice(0, 5) ^ slice(9, 4, -1) ^ slice(4, 6)
        assert_allclose(c, 1)

        assert rmps[0].data.ndim == (3 if cyclic else 2)
        assert rmps[-1].data.ndim == (3 if cyclic else 2)

    def test_mps_computation_state(self):
        p = MPS_neel_state(10)
        pd = qu.neel_state(10)
        assert_allclose(p.to_dense(), pd)

    def test_zero_state(self):
        z = MPS_zero_state(21, 7)
        p = MPS_rand_state(21, 13)
        assert_allclose(p.H @ z, 0.0)
        assert_allclose(p.H @ p, 1.0)
        zp = z + p
        assert max(zp[13].shape) == 20
        assert_allclose(zp.H @ p, 1.0)

    @pytest.mark.parametrize("cyclic", [False, True])
    @pytest.mark.parametrize("j", [7 / 11, 1, (0.2, 0.3, 0.4)])
    @pytest.mark.parametrize("bz", [0, 7 / 11, 1])
    @pytest.mark.parametrize("n", [2, 3, 4])
    def test_mpo_site_ham_heis(self, cyclic, j, bz, n):
        hh_mpo = MPO_ham_heis(n, tags=['foo'], cyclic=cyclic, j=j, bz=bz)
        assert hh_mpo[0].tags == oset(['I0', 'foo'])
        assert hh_mpo[1].tags == oset(['I1', 'foo'])
        assert hh_mpo[-1].tags == oset([f'I{n - 1}', 'foo'])
        assert hh_mpo.shape == (2, ) * 2 * n
        hh_ex = qu.ham_heis(n, cyclic=cyclic, j=j, b=bz)
        assert_allclose(
            qu.eigvalsh(hh_ex), qu.eigvalsh(hh_mpo.to_dense()), atol=1e-13)

    def test_mpo_zeros(self):
        mpo0 = MPO_zeros(10)
        assert mpo0.trace() == 0.0
        assert mpo0.H @ mpo0 == 0.0

    @pytest.mark.parametrize("cyclic", (False, True))
    def test_mpo_zeros_like(self, cyclic):
        A = MPO_rand(10, 7, phys_dim=3, normalize=False, cyclic=cyclic)
        Z = MPO_zeros_like(A)
        assert A @ Z == 0.0
        assert Z.cyclic == cyclic
        x1 = A.trace()
        x2 = (A + Z).trace()
        assert_allclose(x1, x2)


class TestDense1D:

    def test_simple(self):
        n = 10
        d_psi = qu.computational_state('0' * n)

        t_psi = Dense1D(d_psi)
        assert set(t_psi.outer_inds()) == {f'k{i}' for i in range(n)}
        assert t_psi.tags == oset(f'I{i}' for i in range(n))

        for i in range(n):
            assert t_psi.H @ t_psi.gate(qu.pauli('Z'), i) == pytest.approx(1)

        for i in range(n):
            t_psi.gate_(qu.hadamard(), i)

        assert len(t_psi.tensors) == n + 1

        # should have '++++++++++'
        assert t_psi.H @ t_psi == pytest.approx(1)
        for i in range(n):
            assert t_psi.H @ t_psi.gate(qu.pauli('X'), i) == pytest.approx(1)

    def test_rand(self):
        t_psi = Dense1D.rand(7, dtype='complex64')
        assert t_psi.shape == (2,) * 7
        assert t_psi.dtype == 'complex64'
        assert (t_psi.H @ t_psi) == pytest.approx(1.0)



import pytest
import numpy as np
import scipy.sparse as sp
from numpy.testing import assert_allclose

import quimb as qu
from quimb.linalg import SLEPC4PY_FOUND
from quimb.linalg.base_linalg import _rel_window_to_abs_window

eigs_backends = ["auto", "numpy", "scipy"]
svds_backends = ["numpy", "scipy"]

if SLEPC4PY_FOUND:
    eigs_backends += ["slepc-nompi", "slepc"]
    svds_backends += ["slepc-nompi", "slepc"]


# --------------------------------------------------------------------------- #
#                              Fixtures                                       #
# --------------------------------------------------------------------------- #

@pytest.fixture
def mat_herm_dense():
    np.random.seed(1)
    u = qu.rand_uni(4)
    a = u @ qu.ldmul(np.array([-1, 2, 4, -3]), u.H)
    #  |--|--|--|--|--|--|--|
    # -3    -1        2     4
    return u, a


@pytest.fixture
def mat_herm_sparse():
    np.random.seed(1)
    u = qu.rand_uni(4)
    a = u @ qu.ldmul(np.array([-1, 2, 4, -3]), u.H)
    a = qu.sparse(a)
    return u, a


@pytest.fixture
def mat_nherm_dense():
    np.random.seed(1)
    u, v = qu.rand_uni(5), qu.rand_uni(5)
    a = u @ qu.ldmul(np.array([1, 2, 4, 3, 0.1]), v.H)
    return u, v, a


@pytest.fixture
def mat_nherm_sparse():
    np.random.seed(1)
    u, v = qu.rand_uni(5), qu.rand_uni(5)
    a = u @ qu.ldmul(np.array([1, 2, 4, 3, 0.1]), v.H)
    a = qu.sparse(a)
    return u, v, a


@pytest.fixture
def ham1():
    u = qu.rand_uni(7)
    el = np.array([-3, 0, 1, 2, 3, 4, 7])
    return u @ qu.ldmul(el, u.H)


@pytest.fixture
def ham2():
    u = qu.rand_uni(7)
    el = np.array([-3.72, 0, 1, 1.1, 2.1, 2.2, 6.28])
    return u @ qu.ldmul(el, u.H)


# --------------------------------------------------------------------------- #
#                              Tests                                          #
# --------------------------------------------------------------------------- #

class TestEigh:
    def test_eigsys(self, mat_herm_dense):
        u, a = mat_herm_dense
        evals, v = qu.eigh(a)
        assert(set(np.rint(evals)) == set((-1, 2, 4, -3)))
        assert_allclose(evals, [-3, -1, 2, 4])
        for i, j in zip([3, 0, 1, 2], range(4)):
            o = u[:, [i]].H @ v[:, [j]]
            assert_allclose(abs(o), 1.)

    def test_eigvals(self, mat_herm_dense):
        _, a = mat_herm_dense
        evals = qu.eigvalsh(a)
        assert_allclose(evals, [-3, -1, 2, 4])

    def test_eigvecs(self, mat_herm_dense):
        u, a = mat_herm_dense
        v = qu.eigvecsh(a)
        for i, j in zip([3, 0, 1, 2], range(4)):
            o = u[:, [i]].H @ v[:, [j]]
            assert_allclose(abs(o), 1.)


class TestSeigs:
    @pytest.mark.parametrize("backend", eigs_backends)
    def test_eigs_small_dense_wvecs(self, mat_herm_dense, backend):
        u, a = mat_herm_dense
        assert not qu.issparse(a)
        lk, vk = qu.eigh(a, k=2, backend=backend)
        assert_allclose(lk, (-3, -1))
        for i, j in zip([3, 0], [0, 1]):
            o = u[:, [i]].H @ vk[:, [j]]
            assert_allclose(abs(o), 1.)
        vk = qu.eigvecsh(a, k=2, backend=backend)
        for i, j in zip([3, 0], [0, 1]):
            o = u[:, [i]].H @ vk[:, [j]]
            assert_allclose(abs(o), 1.)

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_eigs_small_dense_novecs(self, mat_herm_dense, backend):
        _, a = mat_herm_dense
        assert not qu.issparse(a)
        lk = qu.eigvalsh(a, k=2, backend=backend)
        assert_allclose(lk, (-3, -1))

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_eigs_sparse_wvecs(self, mat_herm_sparse, backend):
        u, a = mat_herm_sparse
        assert qu.issparse(a)
        lk, vk = qu.eigh(a, k=2, backend=backend)
        assert_allclose(lk, (-3, -1))
        for i, j in zip([3, 0], [0, 1]):
            o = u[:, [i]].H @ vk[:, [j]]
            assert_allclose(abs(o), 1.)
        vk = qu.eigvecsh(a, k=2, backend=backend)
        for i, j in zip([3, 0], [0, 1]):
            o = u[:, [i]].H @ vk[:, [j]]
            assert_allclose(abs(o), 1.)

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_eigs_small_sparse_novecs(self, mat_herm_sparse, backend):
        _, a = mat_herm_sparse
        assert qu.issparse(a)
        lk = qu.eigvalsh(a, k=2, backend=backend)
        assert_allclose(lk, (-3, -1))

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_groundstate(self, mat_herm_dense, backend):
        u, a = mat_herm_dense
        gs = qu.groundstate(a, backend=backend)
        assert_allclose(abs(u[:, [3]].H @ gs), 1.)

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_groundenergy(self, mat_herm_dense, backend):
        _, a = mat_herm_dense
        ge = qu.groundenergy(a, backend=backend)
        assert_allclose(ge, -3)

    @pytest.mark.parametrize("which", [None, "SA", "LA", "LM", "SM", "TR"])
    @pytest.mark.parametrize("k", [1, 2])
    def test_cross_equality(self, mat_herm_sparse, k, which):
        _, a = mat_herm_sparse
        sigma = 1 if which in {None, "TR"} else None
        lks, vks = zip(*(qu.eigh(a, k=k, which=which, sigma=sigma, backend=b)
                         for b in eigs_backends))
        lks, vks = tuple(lks), tuple(vks)
        for i in range(len(lks) - 1):
            assert_allclose(lks[i], lks[i + 1])
            assert_allclose(abs(vks[i].H @ vks[i + 1]), qu.eye(k), atol=1e-14)


class TestLOBPCG:
    def test_against_arpack(self):
        A = qu.rand_herm(32, dtype=float)
        lk, vk = qu.eigh(A, k=6, backend='lobpcg')
        slk, svk = qu.eigh(A, k=6, backend='scipy')
        assert_allclose(lk, slk)
        assert_allclose(np.eye(6), abs(vk.H @ svk), atol=1e-9, rtol=1e-9)


class TestEvalsWindowed:
    @pytest.mark.parametrize("backend", eigs_backends)
    def test_bound_spectrum(self, ham1, backend):
        h = ham1
        lmin, lmax = qu.bound_spectrum(h, backend=backend)
        assert_allclose((lmin, lmax), (-3, 7), atol=1e-13)

    def test_rel_window_to_abs_window(self):
        el0 = _rel_window_to_abs_window(5, 10, 0.5)
        assert_allclose(el0, 7.5)
        el0, eli, elf = _rel_window_to_abs_window(-20, -10, 0.5, 0.2)
        assert_allclose([el0, eli, elf], [-15, -16, -14])

    def test_dense(self, ham2):
        h = ham2
        el = qu.eigvalsh_window(h, 0.5, 2, w_sz=0.1)
        assert_allclose(el, [1, 1.1])

    def test_dense_cut(self, ham1):
        h = ham1
        el = qu.eigvalsh_window(h, 0.5, 5, w_sz=0.3)
        assert_allclose(el, [1, 2, 3])

    @pytest.mark.parametrize("backend", eigs_backends)
    def test_sparse(self, ham2, backend):
        h = qu.sparse(ham2)
        el = qu.eigvalsh_window(h, 0.5, 2, w_sz=0.1, backend=backend)
        assert_allclose(el, [1, 1.1])

    def test_sparse_cut(self, ham1):
        h = qu.sparse(ham1)
        el = qu.eigvalsh_window(h, 0.5, 5, w_sz=0.3)
        assert_allclose(el, [1, 2, 3])

    def test_dense_return_vecs(self, mat_herm_dense):
        u, a = mat_herm_dense
        ev = qu.eigvecsh_window(a, w_0=0.5, k=2, w_sz=0.8)
        assert ev.shape == (4, 2)
        assert_allclose(abs(u[:, :2].H @ ev[:, ]), [[1, 0], [0, 1]],
                        atol=1e-14)

    def test_sparse_return_vecs(self, mat_herm_sparse):
        u, a = mat_herm_sparse
        ev = qu.eigvecsh_window(a, w_0=0.5, k=2, w_sz=0.8)
        assert ev.shape == (4, 2)
        assert_allclose(abs(u[:, :2].H @ ev[:, ]), [[1, 0], [0, 1]],
                        atol=1e-14)


class TestSVD:
    def test_svd_full(self, mat_nherm_dense):
        u, v, a = mat_nherm_dense
        un, sn, vn = qu.svd(a)
        assert_allclose(sn, [4, 3, 2, 1, 0.1], atol=1e-14)
        for i, j, in zip((0, 1, 2, 3, 4),
                         (2, 3, 1, 0, 4)):
            o = abs(un[:, [i]].H @ u[:, [j]])
            assert_allclose(o, 1.)
            o = abs(vn[[i], :] @ v[:, [j]])
            assert_allclose(o, 1.)


class TestSVDS:
    @pytest.mark.parametrize("backend", svds_backends)
    def test_svds_smalldense_wvecs(self, mat_nherm_dense, backend):
        u, v, a = mat_nherm_dense
        uk, sk, vk = qu.svds(a, k=3, return_vecs=True, backend=backend)
        assert_allclose(sk, [4, 3, 2])
        for i, j in zip((0, 1, 2), (2, 3, 1)):
            o = abs(uk[:, [i]].H @ u[:, [j]])
            assert_allclose(o, 1.)
            o = abs(vk[[i], :] @ v[:, [j]])
            assert_allclose(o, 1.)

    @pytest.mark.parametrize("backend", svds_backends)
    def test_svds_smalldense_nvecs(self, mat_nherm_dense, backend):
        _, _, a = mat_nherm_dense
        sk = qu.svds(a, k=3, return_vecs=False, backend=backend)
        assert_allclose(sk, [4, 3, 2])

    @pytest.mark.parametrize("backend", svds_backends)
    def test_svds_sparse_wvecs(self, mat_nherm_sparse, backend):
        u, v, a = mat_nherm_sparse
        uk, sk, vk = qu.svds(a, k=3, return_vecs=True, backend=backend)
        assert_allclose(sk, [4, 3, 2])
        for i, j in zip((0, 1, 2), (2, 3, 1)):
            o = abs(uk[:, [i]].H @ u[:, [j]])
            assert_allclose(o, 1.)
            o = abs(vk[[i], :] @ v[:, [j]])
            assert_allclose(o, 1.)

    @pytest.mark.parametrize("backend", svds_backends)
    def test_svds_sparse_nvecs(self, mat_nherm_sparse, backend):
        _, _, a = mat_nherm_sparse
        sk = qu.svds(a, k=3, return_vecs=False, backend=backend)
        assert_allclose(sk, [4, 3, 2])


class TestNorms:
    def test_norm_fro_dense(self):
        a = qu.quimbify([[1, 2], [3j, 4j]])
        assert qu.norm(a, "fro") == (1 + 4 + 9 + 16)**0.5

    def test_norm_fro_sparse(self):
        a = qu.sparse([[3, 0], [4j, 0]])
        assert qu.norm(a, "fro") == (9 + 16)**0.5

    @pytest.mark.parametrize("backend", svds_backends)
    def test_norm_spectral_dense(self, mat_nherm_dense, backend):
        _, _, a = mat_nherm_dense
        assert_allclose(qu.norm(a, "spectral", backend=backend), 4.)

    @pytest.mark.parametrize("backend", svds_backends)
    def test_norm_spectral_sparse(self, mat_nherm_sparse, backend):
        _, _, a = mat_nherm_sparse
        assert_allclose(qu.norm(a, "spectral", backend=backend), 4.)

    def test_norm_trace_dense(self):
        a = qu.qarray(np.diag([-3, 1, 7]))
        assert qu.norm(a, "trace") == 11
        a = qu.rand_product_state(1, qtype="dop")
        assert_allclose(qu.norm(a, "nuc"), 1)


class TestExpm:
    @pytest.mark.parametrize("herm", [True, False])
    def test_zeros_dense(self, herm):
        p = qu.expm(np.zeros((2, 2), dtype=complex), herm=herm)
        assert_allclose(p, qu.eye(2))

    @pytest.mark.parametrize("sparse", [True, False])
    @pytest.mark.parametrize("herm", [True, False])
    def test_eye(self, sparse, herm):
        p = qu.expm(qu.eye(2, sparse=sparse), herm=herm)
        assert_allclose((p.A if sparse else p) / np.e, qu.eye(2))
        if sparse:
            assert isinstance(p, sp.csr_matrix)


class TestSqrtm:
    @pytest.mark.parametrize("sparse", [True, False])
    @pytest.mark.parametrize("herm", [True, False])
    def test_eye(self, herm, sparse):
        if sparse:
            with pytest.raises(NotImplementedError):
                p = qu.sqrtm(qu.eye(2, sparse=sparse), herm=herm)
        else:
            p = qu.sqrtm(qu.eye(2), herm=herm)
            assert_allclose(p, qu.eye(2))


class TestLazy:

    @pytest.mark.parametrize("sparse", [False, True])
    def test_basic(self, sparse):
        ownership = (0, 7)
        hl = qu.Lazy(qu.ham_heis, n=4, sparse=sparse, shape=(16, 16))
        print(hl)
        h = 1 * hl(ownership=ownership)
        h_ex = qu.ham_heis(n=4, sparse=sparse)[slice(*ownership), :]
        assert_allclose(h.A, h_ex.A)

    @pytest.mark.parametrize("backend", ['scipy', 'lobpcg'])
    def test_project_eig(self, backend):
        Hl = qu.Lazy(qu.ham_heis, 4, sparse=True, shape=(16, 16), cyclic=True)
        Pl = qu.Lazy(qu.zspin_projector, 4, shape=(16, 6))

        ge, gs = qu.eigh(Hl, P=Pl, k=1, backend=backend)

        assert ge == pytest.approx(-2)
        assert qu.expec(gs, gs) == pytest.approx(1.0)



"""
Functions for generating quantum states.
"""

import itertools
import functools
import math
import numpy as np

from ..core import (dag, ldmul, dot, make_immutable,
                    qu, kron, eye, ikron, kronpow)
from ..linalg.base_linalg import eigh
from .operators import pauli, controlled


def basis_vec(i, dim, ownership=None, **kwargs):
    """Constructs a unit vector ket.

    Parameters
    ----------
    i : int
        Which index should the single non-zero, unit entry.
    dim : int
        Total size of hilbert space.
    sparse : bool, optional
        Return vector as sparse matrix.
    kwargs
        Supplied to ``qu``.

    Returns:
    --------
    vector
        The basis vector.
    """
    if ownership is None:
        shape = (dim, 1)
        x = np.zeros(shape, dtype=complex)
        x[i] = 1.0
    else:
        ri, rf = ownership
        shape = (rf - ri, 1)
        x = np.zeros(shape, dtype=complex)
        if ri <= i < rf:
            x[i - ri] = 1.0

    return qu(x, **kwargs)


@functools.lru_cache(8)
def up(**kwargs):
    """Returns up-state, aka. ``|0>``, +Z eigenstate.
    """
    u = qu([[1], [0]], **kwargs)
    make_immutable(u)
    return u


zplus = up


@functools.lru_cache(8)
def down(**kwargs):
    """Returns down-state, aka. ``|1>``, -Z eigenstate.
    """
    d = qu([[0], [1]], **kwargs)
    make_immutable(d)
    return d


zminus = down


@functools.lru_cache(8)
def plus(**kwargs):
    """Returns plus-state, aka. ``|+>``, +X eigenstate.
    """
    return qu([[2**-0.5], [2**-0.5]], **kwargs)


xplus = plus


@functools.lru_cache(8)
def minus(**kwargs):
    """Returns minus-state, aka. ``|->``, -X eigenstate.
    """
    return qu([[2**-0.5], [-2**-0.5]], **kwargs)


xminus = minus


@functools.lru_cache(8)
def yplus(**kwargs):
    """Returns yplus-state, aka. ``|y+>``, +Y eigenstate.
    """
    return qu([[2**-0.5], [1.0j / (2**0.5)]], **kwargs)


@functools.lru_cache(8)
def yminus(**kwargs):
    """Returns yplus-state, aka. ``|y->``, -Y eigenstate.
    """
    return qu([[2**-0.5], [-1.0j / (2**0.5)]], **kwargs)


def bloch_state(ax, ay, az, purified=False, **kwargs):
    """Construct qubit density operator from bloch vector.

    Parameters
    ----------
    ax : float
        X component of bloch vector.
    ay : float
        Y component of bloch vector.
    az : float
        Z component of bloch vector.
    purified :
        Whether to map vector to surface of bloch sphere.

    Returns
    -------
    Matrix
        Density operator of qubit 'pointing' in (ax, ay, az) direction.
    """
    n = (ax**2 + ay**2 + az**2)**.5
    if purified:
        ax, ay, az = (a / n for a in (ax, ay, az))
    return sum(0.5 * a * pauli(s, **kwargs)
               for a, s in zip((1, ax, ay, az), "ixyz"))


@functools.lru_cache(maxsize=8)
def bell_state(s, **kwargs):
    r"""One of the four bell-states.

    If n = 2**-0.5, they are:

        0. ``'psi-'`` : ``n * ( |01> - |10> )``
        1. ``'psi+'`` : ``n * ( |01> + |10> )``
        2. ``'phi-'`` : ``n * ( |00> - |11> )``
        3. ``'phi+'`` : ``n * ( |00> + |11> )``

    They can be enumerated in this order.

    Parameters
    ----------
    s : str or int
        String of number of state corresponding to above.
    kwargs :
        Supplied to ``qu`` called on state.

    Returns
    -------
    p : immutable vector
        The bell-state ``s``.
    """
    keymap = {"psi-": "psi-", 0: "psi-", "psim": "psi-",
              "psi+": "psi+", 1: "psi+", "psip": "psi+",
              "phi-": "phi-", 2: "phi-", "phim": "phi-",
              "phi+": "phi+", 3: "phi+", "phip": "phi+"}
    c = 2.**-.5
    statemap = {"psi-": lambda: qu([0, c, -c, 0], **kwargs),
                "phi+": lambda: qu([c, 0, 0, c], **kwargs),
                "phi-": lambda: qu([c, 0, 0, -c], **kwargs),
                "psi+": lambda: qu([0, c, c, 0], **kwargs)}
    state = statemap[keymap[s]]()
    make_immutable(state)
    return state


def singlet(**kwargs):
    """Alias for the 'psi-' bell-state.
    """
    return bell_state("psi-", **kwargs)


def thermal_state(ham, beta, precomp_func=False):
    """Generate a thermal state of a Hamiltonian.

    Parameters
    ----------
    ham : operator or (1d-array, 2d-array)
        Hamiltonian, either full or tuple of (evals, evecs).
    beta : float
        Inverse temperature of state.
    precomp_func : bool, optional
        If True, return a function that takes ``beta``
        only and is closed over the solved hamiltonian.

    Returns
    -------
    operator or callable
        Density operator of thermal state, or function to generate such given
        a temperature.
    """
    if isinstance(ham, (list, tuple)):  # solved already
        evals, evecs = ham
    else:
        evals, evecs = eigh(ham)
    evals -= evals.min()  # offset by min to avoid numeric problems

    def gen_state(b):
        el = np.exp(-b * evals)
        el /= np.sum(el)
        return dot(evecs, ldmul(el, dag(evecs)))

    return gen_state if precomp_func else gen_state(beta)


def computational_state(binary, **kwargs):
    """Generate the qubit computational state with ``binary``.

    Parameters
    ----------
    binary : sequence of 0s and 1s
        The binary of the computation state.

    Examples
    --------
    >>> computational_state('101'):
    qarray([[0.+0.j],
            [0.+0.j],
            [0.+0.j],
            [0.+0.j],
            [0.+0.j],
            [1.+0.j],
            [0.+0.j],
            [0.+0.j]])

    >>> qu.computational_state([0, 1], qtype='dop')
    qarray([[0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
            [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]])

    See Also
    --------
    MPS_computational_state, basic_vec
    """
    if not isinstance(binary, str):
        binary = "".join(map(str, binary))

    return basis_vec(int(binary, 2), 2 ** len(binary), **kwargs)


def neel_state(n, down_first=False, **kwargs):
    """Construct Neel state for n spins, i.e. alternating up/down.

    Parameters
    ----------
    n : int
        Number of spins.
    down_first : bool, optional
        Whether to start with '1' or '0' first.
    kwargs
        Supplied to ``qu`` called on state.

    See Also
    --------
    computational_state, MPS_neel_state
    """
    binary = "01" * (n // 2) + (n % 2 == 1) * "0"
    if down_first:
        binary = "1" + binary[:-1]

    return computational_state(binary, **kwargs)


def singlet_pairs(n, **kwargs):
    """Construct fully dimerised spin chain.

    I.e. ``bell_state('psi-') & bell_state('psi-') & ...``

    Parameters
    ----------
    n : int
        Number of spins.
    kwargs
        Supplied to ``qu`` called on state.

    Returns
    -------
    vector
    """
    return kronpow(bell_state('psi-', **kwargs), (n // 2))


def werner_state(p, **kwargs):
    """Construct Werner State, i.e. fractional mix of identity with singlet.

    Parameters
    ----------
    p : float
        Singlet Fraction.
    kwargs
        Supplied to :func:`~quimb.qu` called on state.

    Returns
    -------
    qarray
    """
    return (p * bell_state('psi-', qtype="dop", **kwargs) +
            (1 - p) * eye(4, **kwargs) / 4)


def ghz_state(n, **kwargs):
    """Construct GHZ state of `n` spins, i.e. equal superposition of all up
    and down.

    Parameters
    ----------
    n : int
        Number of spins.
    kwargs
        Supplied to ``qu`` called on state.

    Returns
    -------
    vector
    """
    psi = (basis_vec(0, 2**n) +
           basis_vec(2**n - 1, 2**n)) / 2.**.5
    return qu(psi, **kwargs)


def w_state(n, **kwargs):
    """Construct W-state: equal superposition of all single spin up states.

    Parameters
    ----------
    n : int
        Number of spins.
    kwargs
        Supplied to ``qu`` called on state.

    Returns
    -------
    vector
    """
    return sum(basis_vec(2**i, 2**n, **kwargs) for i in range(n)) / n**0.5


def levi_civita(perm):
    """Compute the generalised levi-civita coefficient for a permutation.

    Parameters
    ----------
    perm : sequence of int
        The permutation, a re-arrangement of ``range(n)``.

    Returns
    -------
    int
        Either -1, 0 or 1.
    """
    n = len(perm)
    if n != len(set(perm)):  # infer there are repeated elements
        return 0
    mat = np.zeros((n, n), dtype=np.int32)
    for i, j in zip(range(n), perm):
        mat[i, j] = 1
    return int(np.linalg.det(mat))


def perm_state(ps):
    """Construct the anti-symmetric state which is the +- sum of all
    tensored permutations of states ``ps``.

    Parameters
    ----------
    ps :  sequence of states
        The states to combine.

    Returns
    -------
    vector or operator
        The permutation state, dimension same as ``kron(*ps)``.

    Examples
    --------
    A singlet is the ``perm_state`` of up and down.

    >>> states = [up(), down()]
    >>> pstate = perm_state(states)
    >>> expec(pstate, singlet())
    1.0
    """
    n = len(ps)
    vec_perm = itertools.permutations(ps)
    ind_perm = itertools.permutations(range(n))

    def terms():
        for vec, ind in zip(vec_perm, ind_perm):
            yield levi_civita(ind) * kron(*vec)

    return sum(terms()) / math.factorial(n)**0.5


def graph_state_1d(n, cyclic=False, sparse=False):
    """Graph State on a line.

    Parameters
    ----------
    n : int
        The number of spins.
    cyclic : bool, optional
        Whether to use cyclic boundary conditions for the graph.
    sparse : bool, optional
        Whether to return a sparse state.

    Returns
    -------
    vector
        The 1d-graph state.
    """
    p = kronpow(plus(sparse=sparse), n)

    for i in range(n - 1):
        p = ikron(controlled("z", sparse=True), [2] * n, (i, i + 1)) @ p

    if cyclic:
        p = ((eye(2, sparse=True) & eye(2**(n - 2), sparse=True) &
              qu([1, 0], qtype="dop", sparse=True)) +
             (pauli("z", sparse=True) & eye(2**(n - 2), sparse=True) &
              qu([0, 1], qtype="dop", sparse=True))) @ p

    return p
